# -*- coding: utf-8 -*-
"""
@file
@brief Defines a set of modules for more machine learning or student projects.
"""
import sys
from ..installhelper.module_install import ModuleInstall


def ml_set():
    """
    modules introduced by students or needed for student projects
    """
    return ensae_set()


def ensae_set():
    """
    modules introduced by students or needed for student projects, it requires the modules in set *extended*
    """
    mod = [
        ModuleInstall(
            "billiard", "pip", purpose="Python multiprocessing fork with improvements and bugfixes (for celery)"),
        ModuleInstall(
            "vine", "pip", purpose="Promises, promises, promises"),
        ModuleInstall(
            "amqp", "pip", purpose="Low-level AMQP client for Python (fork of amqplib) (for celery)"),
        ModuleInstall(
            "anyjson", "pip", purpose="Wraps the best available JSON implementation available in a common interface (for celery)"),
        ModuleInstall(
            "kombu", "pip", purpose="Messaging library for Python (for celery)"),
        ModuleInstall(
            "celery", "pip", purpose="Celery is an asynchronous task queue/job queue based on distributed message passing."),

        ModuleInstall('html2text', 'pip',
                      purpose="Turn HTML into equivalent Markdown-structured text."),
        ModuleInstall('easy-thumbnails', 'pip', mname="easy_thumbnails",
                      purpose="Easy thumbnails for Django"),
        ModuleInstall('Versio', 'pip', mname="versio",
                      purpose="localshop dependency, manages versions"),
        ModuleInstall('django', 'pip',
                      purpose="Django"),
        ModuleInstall('django-celery', 'pip', mname="djcelery",
                      purpose="localshop dependency, Old django celery integration project."),
        ModuleInstall('django-configurations', 'pip', mname="configurations",
                      purpose="localshop dependency, A helper for organizing Django settings."),
        ModuleInstall('django-environ', 'pip', mname="environ",
                      purpose="localshop dependency, Django-environ allows you to utilize 12factor "
                      + "inspired environment variables to configure your Django application."),
        ModuleInstall('django-model-utils', 'pip', mname="model_utils",
                      purpose="localshop dependency, Django model mixins and utilities."),
        ModuleInstall('django-storages', 'pip', mname="storages",
                      purpose="localshop dependency, django-storages is a collection of custom storage backends for Django."),
        ModuleInstall('django-guardian', 'pip', mname="guardian",
                      purpose="localshop dependency, Implementation of per object permissions for Django 1.2 or later."),
        ModuleInstall('django-userena', 'pip', mname="userena",
                      purpose="localshop dependency, Accounts for Django made beautifully simple"),
        ModuleInstall('django-uuidfield', 'pip', mname="uuidfield",
                      purpose="localshop dependency, UUIDField in Django"),
        ModuleInstall("grappelli_safe", "pip", usage="WEB",
                      purpose="A snapshot of the grappelli_2 branch of django-grappelli, packaged as a dependency "
                      + "for the Mezzanine CMS for Django."),
        ModuleInstall("filebrowser_safe", "pip", usage="WEB",
                      purpose="A snapshot of the filebrowser_3 branch of django-filebrowser, packaged as a "
                      + "dependency for the Mezzanine CMS for Django."),
        ModuleInstall("django-contrib-comments", "pip", usage="WEB", mname="django_comments",
                      purpose="Django used to include a comments framework; since Django 1.6 it’s "
                      + "been separated to a separate project. This is that project."),
        ModuleInstall("mezzanine", "pip", usage="WEB",
                      purpose="Mezzanine is a powerful, consistent, and flexible content management platform."),
        ModuleInstall(
            'gunicorn', 'pip', purpose="localshop dependency, WSGI HTTP Server for UNIX"),
        ModuleInstall(
            'netaddr', 'pip', purpose="Pythonic manipulation of IPv4, IPv6, CIDR, EUI and MAC network addresses"),
        # ModuleInstall('localshop', 'pip',
        # purpose="run a local pypi server (install it in virtual env as it
        # overwrites many versions)"),

        ModuleInstall(
            "tweepy", "pip", purpose="Python wrapper for the twitter API"),
        #ModuleInstall("newspaper3k", "pip", mname="newspaper"),
        ModuleInstall(
            "mutagen", "pip", purpose="read and write audio tags for many formats in Python 3"),
        ModuleInstall(
            "mutagenx", "pip", purpose="read and write audio tags for many formats in Python 3"),
        ModuleInstall("django-audiotracks", "pip",
                      mname="audiotracks", purpose="read audio with django"),
        ModuleInstall("ndg-httpsclient", "pip", mname="ndg.httpsclient",
                      purpose="Provides enhanced HTTPS support for httplib and urllib2 using PyOpenSSL"),
        ModuleInstall("inflection", "pip",
                      purpose="A port of Ruby on Rails inflector to Python"),
        ModuleInstall("more-itertools", "pip",
                      mname="more_itertools",
                      purpose="More routines for operating on iterables, beyond itertools"),
        ModuleInstall("Quandl", "pip", mname="quandl",
                      purpose="access Quandl API"),
        ModuleInstall("singledispatch", "pip", purpose="for nltk"),
        ModuleInstall(
            "nltk", "pip", purpose="NLP, natural language processing", usage="NLP"),
        ModuleInstall(
            "textblob", "pip", purpose="TextBlob is a Python (2 and 3) library for processing textual data. It provides a simple API for "
            + "diving into common natural language processing (NLP) tasks such as part-of-speech tagging, "
            + "noun phrase extraction, sentiment analysis, classification, translation, and more."),
        ModuleInstall(
            "opencv_python", "wheel", mname="cv2", purpose="OpenVC wrapper",
            web="https://opencv-python-tutroals.readthedocs.org/en/latest/"),
        ModuleInstall("dlib", "pip" if sys.version_info[:2] >= (3, 6) else "wheel",
                      source="2" if sys.version_info[:2] >= (3, 6) else None,
                      usage="ML",
                      purpose="A toolkit for making real world machine learning and data analysis applications"),
        ModuleInstall("PyAudio", "wheel", mname="pyaudio",
                      purpose="PyAudio provides Python bindings for PortAudio v19, the cross-platform audio I/O library. "
                      + "With PyAudio, you can easily use Python to play and record audio "
                      + "streams on a variety of platforms (e.g., GNU/Linux, Microsoft Windows, and Mac OS X)."),
        ModuleInstall(
            "zope.interface", "wheel", purpose="interfaces for python"),
        ModuleInstall(
            "zope.exceptions", "pip", purpose="Zope exception"),
        ModuleInstall(
            "persistent", "wheel", purpose="Objets persistants translucides"),
        # requires zope.interface, persistents
        ModuleInstall("BTrees", "wheel", usage="ALGO",
                      purpose="This package contains a set of persistent object containers built around a modified "
                      + "BTree data structure."),
        ModuleInstall(
            "datrie", "wheel", purpose="Fast, efficiently stored Trie for Python.", usage="ALGO"),
        # ModuleInstall("pysparse", "pip"), #does not work
        ModuleInstall(
            "mahotas", "wheel", purpose="Mahotas: Computer Vision Library", usage="VIZ"),
        ModuleInstall("nitime", "wheel",
                      purpose="Nitime is a library for time-series analysis of data from neuroscience experiments.", usage="DATA/ML"),
        ModuleInstall("milk", "wheel",
                      purpose="machine learning toolkit", usage="DATA/ML"),
        ModuleInstall("minepy", "wheel", purpose="interface to MineCraft"),
        ModuleInstall(
            "NLopt", "wheel", mname="nlopt", purpose="linear, quadratic optimization",
            web="http://ab-initio.mit.edu/wiki/index.php/NLopt", usage="DATA/ML"),
        ModuleInstall("Pmw", "pip",
                      purpose="Pmw is a toolkit for building high-level compound widgets in Python using the Tkinter module."),
        ModuleInstall(
            "pytool", "pip", purpose="A collection of tools for Python"),
        ModuleInstall(
            "pytools", "pip", purpose="A collection of tools for Python"),
        ModuleInstall("pycuda", "wheel", usage="GPU",
                      purpose="PyCUDA lets you access Nvidia's CUDA parallel computation API from Python."),
        ModuleInstall("pyopencl", "wheel", usage="GPU",
                      purpose="PyOpenCL lets you access the OpenCL parallel computation API from Python."),
        # ModuleInstall("scikits.cuda", "pip", mname="skcuda"), # no stable
        # version
        ModuleInstall(
            "pylzma", "wheel2", purpose="Python bindings for the LZMA library by Igor Pavlov."),
        ModuleInstall("nibabel", "pip",
                      purpose="Access a multitude of neuroimaging data formats."),
        ModuleInstall(
            "pyodbc", "wheel", purpose="access to protocal ODBC (SQL databases)", usage="SQL"),
        ModuleInstall(
            "pypmc", "wheel", purpose="pypmc is a python package focusing on adaptive importance sampling."),
        ModuleInstall("PyX", "wheel", mname="pyx",
                      purpose="plotting", usage="VIZ"),
        ModuleInstall(
            "scandir", "wheel", purpose="Better directory iterator and faster os.walk(), "
            + "now in the Python 3.5 stdlib") if sys.version_info[:2] <= (3, 4) else None,
        ModuleInstall(
            "backports.lzma", "wheel", purpose="Backport of Python 3.3's 'lzma' module for XZ/LZMA compressed files."),
        ModuleInstall(
            "zs", "wheel", purpose="S is a compressed, read-only file format for efficiently distributing, "
            + "querying, and archiving arbitrarily large record-oriented datasets."),
        #
        # teachings
        #
        ModuleInstall(
            "tutormagic", "pip", purpose="brings PythonTutor in a notebok", usage="TEACH"),
        # to upload a file in a notebook
        ModuleInstall(
            "nbupload", "pip", purpose="widget to upload a file in a notebook",
            usage="JUPYTER"),
        # see https://github.com/PetterS/numpy_display/blob/master/numpy_display.py
        # https://github.com/damiendr/callipy
        #
        ModuleInstall("libsvm", "wheel", mname="svm", purpose="SVM library"),
        # Bayesian ABC
        ModuleInstall("abcpmc", "pip", purpose="Monte Carlo and ABC methods"),
        # ModuleInstall("cosmoabc", "pip"), # Bayesian ABC, only python 2.7
        #
        # ModuleInstall("pyjs", "pip"), # needs manual installation
        # ModuleInstall("pyjs", "github", "pyjs"), #does not work really
        # ModuleInstall("pyjsdl", "github", "jggatc"), # no setup.py

        #
        #
        #
        # ModuleInstall("contextlib2", "pip", purpose="Backports and enhancements for the contextlib module"),
        ModuleInstall(
            "nuitka", "pip", usage="C++", purpose="C++ compilation, code optimization"),
        # ModuleInstall("tri", "pip", purpose="Delaunay triangulation"), # only
        # works on Python 2.7
        ModuleInstall(
            "blosc", "wheel", purpose="Blosc (http://blosc.org) is a high performance compressor optimized for binary data."),
        ModuleInstall(
            "tables", "wheel", purpose="PyTables is a package for managing hierarchical datasets "
            + "and designed to efficiently and easily cope with extremely large amounts of data."),
        ModuleInstall("contextlib2", "pip",
                      purpose="Backports and enhancements for the contextlib module"),
        ModuleInstall(
            "sortedcontainers", "pip",
            purpose="Python Sorted Container Types: SortedList, SortedDict, and SortedSet"),
        ModuleInstall("python-editor", "pip", mname="python_editor",
                      purpose="Programmatically open an editor, capture the result."),
        ModuleInstall(
            "alembic", "pip",
            purpose="A database migration tool for SQLAlchemy."),
        ModuleInstall(
            "intervaltree", "pip",
            purpose="Editable interval tree data structure for Python 2 and 3"),
        ModuleInstall(
            "cachetools", "pip",
            purpose="Extensible memoizing collections and decorators"),
        ModuleInstall(
            "empyrical", "pip",
            purpose="empyrical is a Python library with performance and risk statistics commonly used in quantitative finance"),
        ModuleInstall(
            "lru_dict", "pip", purpose="An Dict like LRU container."),
        ModuleInstall(
            "zipline", "wheel",
            purpose="A backtester for financial algorithms.") if sys.version_info[:2] >= (3, 5) else None,  # finance
        ModuleInstall("vincent", "pip", purpose="plotting",
                      usage="VIZ"),  # graph
        # graph, pygal_maps_world only accepts the latest version
        #ModuleInstall("pygal", "github", "Kozea", purpose="plotting"),
        ModuleInstall(
            "pygal", "pip", "Kozea", purpose="plotting (javascript)", usage="VIZ"),
        ModuleInstall(
            "pygal_maps_world", "pip", purpose="extension to pygal (maps)", usage="VIZ"),  # graph
        ModuleInstall(
            "pygal_maps_fr", "pip", purpose="French maps for pygal", usage="VIZ"),  # graph
        ModuleInstall(
            "pygal_maps_ch", "pip", purpose="Swiss canton map for pygal", usage="VIZ"),  # graph
        ModuleInstall(
            "pygal_sphinx_directives", "pip", purpose="Pygal sphinx integration", usage="SPHINX"),  # graph
        #
        # 2015-06-30
        #
        ModuleInstall(
            "sas7bdat", "pip", purpose="read/write SAS format"),  # SAS
        #
        # 2015-07-15
        #
        # linear optimisation, see
        # http://blog.yhathq.com/posts/decision-making-under-uncertainty.html
        ModuleInstall("PuLP", "wheel", mname="pulp",
                      purpose="linear, quadratique optimization with constraints", usage="DATA/ML"),
        #
        # pydata
        #
        ModuleInstall("jaraco.functools", "pip",
                      purpose="Additional functools in the spirit of stdlib’s functools."),
        ModuleInstall("tempora", "pip",
                      purpose="Objects and routines pertaining to date and time (tempora)"),
        ModuleInstall("portend", "pip",
                      purpose="TCP port monitoring utilities"),
        ModuleInstall("cheroot", "pip",
                      purpose="Highly-optimized, pure-python HTTP server"),
        ModuleInstall("CherryPy", "pip", mname="cherrypy",
                      purpose="create web application, needed by Spyre"),
        ModuleInstall("dataspyre", "pip", mname="spyre",
                      purpose="create simple web application to visualize data", usage="VIZ"),
        # ModuleInstall("python-recsys", "github", "ocelma", mname="recsys",
        # purpose="recommendation system", usage="DATA/ML"), #only works on
        # Python 2.7 + csc-pysparse + dividi2 (not maintained anymore)
        ModuleInstall(
            "colorspacious", "pip", purpose="A powerful, accurate, and easy-to-use Python library for doing "
            + "colorspace conversions (for viscm)"),
        ModuleInstall(
            "viscm", "pip", purpose="tool for analyzing colormaps and creating new colormaps."),
        # ModuleInstall("cubehelix", "github", "jradavenport",
        #               purpose="a full implementation of Dave Green's cubehelix colormap for Python",
        #              web="https://github.com/jradavenport/cubehelix"),
        ModuleInstall("lifelines", "pip",
                      purpose="survival analysis", usage="OPTIM"),
        # ModuleInstall("pysnptools", "pip", purpose="operation on DNA sequences"), # only available on Python 2.7
        #
        # 2015-07
        #
        ModuleInstall(
            "heatmap", "wheel", purpose="draw heatmap", usage="VIZ"),
        ModuleInstall("planar", "wheel",
                      purpose="2D planar geometry library for Python."),
        ModuleInstall("GDAL", "wheel", mname="osgeo",
                      purpose="GDAL is a translator library for raster and vector geospatial data formats "
                      + "that is released under an X/MIT style "
                      + "Open Source license by the Open Source Geospatial Foundation."),
        # ModuleInstall("rasterio", "wheel",
        # purpose="Fast and direct raster I/O for use with Numpy and SciPy,
        # Rasterio reads and writes geospatial raster datasets."),
        ModuleInstall("cgal_bindings", "wheel", mname="CGAL",
                      purpose="The CGAL Bindings project allows to use some packages of CGAL, the Computational Algorithms Library, "
                      + "in languages other than C++, as for example Java and Python.",
                      web="https://github.com/cgal/cgal-swig-bindings") if sys.version_info[:2] <= (3, 4) else None,
        ModuleInstall("slicerator", "pip",
                      purpose="A lazy-loading, fancy-sliceable iterable."),
        ModuleInstall("PIMS", "pip", mname="pims",
                      purpose="Python Image Sequence (for trackpy)"),
        ModuleInstall("trackpy", "pip",
                      purpose="trackpy is a Python package for particle tracking in 2D, 3D, and higher dimensions.", usage="DATA/ML"),
        ModuleInstall("triangle", "wheel",
                      purpose="Python Triangle is a python wrapper around Jonathan Richard Shewchuk's "
                      + "two-dimensional quality mesh generator and delaunay triangulator library."),
        ModuleInstall("redis", "pip",
                      purpose="Python client for Redis key-value store"),
        # ModuleInstall("vowpal_porpoise", "pip",
        #              purpose="Lightweight python wrapper for vowpal_wabbit.", purpose="DATA/ML"),
        # it requires to build vowpal_wabbit for Windows
        # ModuleInstall("ua-parser", "pip", mname="ua_parser",
        #              purpose="Python port of Browserscope's user agent parser"),
        # ModuleInstall("user-agents", "pip", mname="user_agents",
        # purpose="A library to identify devices (phones, tablets) and their
        # capabilities by parsing (browser/HTTP) user agent strings"),
        ModuleInstall("user-agent", "pip", mname="user_agent",
                      purpose="A library to identify devices (phones, tablets) and their capabilities by "
                      + "parsing (browser/HTTP) user agent strings"),
        ModuleInstall("tinydb", "pip",
                      purpose="TinyDB is a tiny, document oriented database optimized for your happiness :) "
                      + "It's written in pure Python and has no external requirements.", usage="noSQL"),
        ModuleInstall("gevent", "pip",
                      purpose="gevent is a coroutine-based Python networking library"),
        ModuleInstall("grequests", "pip",
                      purpose="GRequests allows you to use Requests with Gevent to make asynchronous HTTP Requests easily."),
        ModuleInstall("pytils", "pip",
                      purpose="Russian-specific string utils (for selection, weblib)"),
        ModuleInstall("selection", "pip",
                      purpose="API to extract data from HTML and XML documents. (for grab)"),
        ModuleInstall("weblib", "pip",
                      purpose="Set of tools for web scraping projects (for grab)"),
        ModuleInstall("grab", "pip",
                      purpose="Grab is a python web scraping framework. Grab provides tons of helpful methods to scrape web sites "),
        ModuleInstall("imbox", "pip",
                      purpose="Python library for reading IMAP mailboxes and converting email content to machine readable data"),
        ModuleInstall("neural-python", "pip", mname="neuralpy",
                      purpose="NeuralPy is the Artificial Neural Network library implemented in Python.", usage="DATA/ML"),
        ModuleInstall("cssselect", "pip",
                      purpose="cssselect parses CSS3 Selectors and translates them to XPath 1.0"),
        ModuleInstall("jieba", "pip",
                      purpose="Chinese Words Segementation Utilities"),
        # goose still depends on BeautifulSoup which does not work on Python 3
        # ModuleInstall("goose-extractor", "pip", mname="goose",
        # purpose="Html Content / Article Extractor, web scrapping lib in
        # Python"),
        ModuleInstall("untangle", "pip",
                      purpose="Converts XML to Python objects"),
        # promising but not released yet
        # ModuleInstall("code2flow", "pip",
        # purpose="Turn your Python and Javascript code into DOT flowcharts"),

        # biokit
        ModuleInstall('colorlog', 'pip',
                      purpose="Log formatting with colors!"),
        ModuleInstall("easydev", "pip",
                      purpose="Common utilities to ease the development of Python packages"),
        ModuleInstall("colormap", "pip", usage="VIZ",
                      purpose="Utilities to ease manipulation of matplotlib colormaps and color codecs (e.g., hex2rgb)"),
        ModuleInstall("suds-jurko", "pip", mname="suds",
                      purpose="Lightweight SOAP client (Jurko's fork)"),
        # bugged and still uses ordereddict
        # ModuleInstall("bioservices", "pip",
        #              purpose="Access to Biological Web Services from Python"),
        # ModuleInstall("biokit", "github", "biokit", usage="VIZ",
        # purpose="Access to Biological Web Services from Python"),

        #
        #
        #
        ModuleInstall("linecache2", "pip",
                      purpose="A backport of linecache to older supported Pythons."),
        ModuleInstall("traceback2", "pip",
                      purpose="traceback2 is a backport of the new features added to the traceback "
                      + "testing framework in Python 2.7 and onwards."),
        ModuleInstall("unittest2", "pip",
                      purpose="unittest2 is a backport of the new features added to the unittest "
                      + "testing framework in Python 2.7 and onwards."),
        ModuleInstall("deprecation", "pip",
                      purpose="A library to handle automated deprecations"),
        ModuleInstall("lz4", "wheel",
                      purpose="LZ4 Bindings for Python (for dpark)"),
        ModuleInstall("fabric", "pip",
                      purpose="Fabric is a Python library and command-line tool for streamlining "
                      + "the use of SSH for application deployment or systems administration tasks."),
        ModuleInstall("invoke", "pip",
                      purpose="Invoke is a Python task execution tool & library, drawing inspiration "
                      + "from various sources to arrive at a powerful & clean feature set."),
        # ModuleInstall("msgpack", "wheel",
        #               purpose="MessagePack (de)serializer."),
        ModuleInstall("msgpack", "wheel",
                      purpose="MessagePack (de)serializer."),
        ModuleInstall("cymem", "pip",
                      purpose="Manage calls to calloc/free through Cython") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall("murmurhash", "pip",
                      purpose="Cython .pxd files for some of the MurmurHash 2 and 3 hash functions, with a slightly more Pythonic API. "
                      + "The only access to these functions is via Cython — I don’t see why they should be "
                      + "useful from pure Python.") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall("preshed", "wheel",
                      purpose="Cython hash table that trusts the keys are pre-hashed") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall("text-unidecode", "pip", mname="text_unidecode",
                      purpose="The most basic Text::Unidecode port") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall("termcolor", "pip",
                      purpose="ANSII Color formatting for output in terminal."),
        ModuleInstall("msgpack-numpy", "pip", mname="msgpack_numpy",
                      purpose="Numpy data serialization using msgpack"),
        ModuleInstall("thinc", "wheel", usage="OPTIM",
                      purpose="Practical Machine Learning for NLP. Thinc is the machine learning "
                      + "library powering spaCy.") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall("semver", "pip",
                      purpose="Python helper for Semantic Versioning (http://semver.org/)"),
        ModuleInstall("sputnik", "pip",
                      purpose="Data package manager library") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall("ftfy", "pip",
                      purpose="Fixes some problems with Unicode text after the fact"),
        ModuleInstall("regex", "wheel",
                      purpose="Alternative regular expression module, to replace re."),
        ModuleInstall("spacy", "wheel", usage="NLP",
                      purpose="Industrial-strength NLP") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall("fr_core_news_sm", "wheel2", usage="NLP",
                      purpose="French ressources for spacy") if sys.version_info[:2] >= (3, 5) else None,

        #
        #  2015-12
        #
        ModuleInstall("picklable-itertools", "pip", mname="picklable_itertools",
                      purpose="A reimplementation of the Python standard library's itertools, in Python, "
                      + "using picklable iterator objects. (fuel)"),
        ModuleInstall("python-utils", "pip", mname="python_utils",
                      purpose="Python Utils is a module with some convenient utilities not included with the standard Python install"),
        ModuleInstall("progressbar2", "pip", usage="VIZ",
                      purpose="Text progress bar library for Python. (fuel)"),
        ModuleInstall("gatspy", "pip",
                      purpose="General tools for Astronomical Time Series in Python"),
        ModuleInstall("supersmoother", "pip",
                      purpose="This is an efficient implementation of Friedman's SuperSmoother [1] algorithm in pure Python. "
                      + "It makes use of numpy for fast numerical computation."),

        #
        # 2016-03
        #
        ModuleInstall("PTable", "pip", mname="prettytable",
                      purpose="A simple Python library for easily displaying tabular data in a visually "
                      + "appealing ASCII table format. (for streamparse)"),
        ModuleInstall("ruamel.yaml", "pip",
                      purpose="ruamel.yaml is a YAML parser/emitter that supports roundtrip preservation of comments, "
                      + "seq/map flow style, and map key order"),
        ModuleInstall("logutils", "pip",
                      purpose="Logging utilities"),
        ModuleInstall("skll", "pip", usage="ML",
                      purpose="SciKit-Learn Laboratory makes it easier to run machinelearning experiments with scikit-learn."),
        ModuleInstall("sklearn_pandas", "pip", usage="ML",
                      purpose="This module provides a bridge between Scikit-Learn's machine learning methods and pandas-style Data Frames."),
        ModuleInstall("ad3", "wheel", source="2", usage="ML",
                      purpose="AD3 (approximate MAP decoder with Alternating Direction Dual Decomposition)"),
        ModuleInstall("pystruct", "wheel", usage="ML", source="2",
                      purpose="Learning Structured Prediction in Python"),
        ModuleInstall("sklearn_contrib_py_earth", "wheel", mname="pyearth", usage="ML", source="2",
                      purpose="A Python implementation of Jerome Friedman's Multivariate Adaptive Regression "
                      + "Splines algorithm, in the style of "
                      + "scikit-learn. The py-earth package implements Multivariate Adaptive Regression Splines using Cython and provides an "
                      + "interface that is compatible with scikit-learn's Estimator, Predictor, Transformer, and Model interfaces. For more "
                      + "information about Multivariate Adaptive Regression Splines, see the references below."),
        ModuleInstall("seqlearn", "wheel", usage="ML",
                      purpose="sequence classification toolkit for Python"),
        ModuleInstall("hmmlearn", "wheel", usage="ML",
                      purpose="Hidden Markov Models in Python, with scikit-learn like API"),
        ModuleInstall("gplearn", "pip", usage="ML",
                      purpose="gplearn implements Genetic Programming in Python, with a scikit-learn inspired and compatible API."),
        ModuleInstall("gdbn", "pip", usage="ML",
                      purpose="This package contains python code for pre-trained deep neural networks"),
        ModuleInstall("gnumpy", "pip",
                      purpose="Gnumpy: an easy way to use GPU boards in Python"),
        ModuleInstall("mlxtend", "pip", usage="ML",
                      purpose="A library consisting of useful tools and extensions for the day-to-day data science tasks."),
        ModuleInstall("jsonpickle", "pip",
                      purpose="Python library for serializing any arbitrary object graph into JSON."),
        ModuleInstall("py-cpuinfo", "pip", mname="py_cpuinfo",
                      purpose="Py-cpuinfo gets CPU info with pure Python. Py-cpuinfo should work without "
                      + "any extra programs or libraries, beyond what your OS provides. It does not "
                      + "require any compilation(C/C++, assembly, et cetera) to use. It works with Python 2 and 3."),
        ModuleInstall("sacred", "pip", usage="ML",
                      purpose="Facilitates automated and reproducible experimental research"),
        ModuleInstall("astropy", "wheel",
                      purpose="Community-developed python astronomy tools"),
        # h2o
        ModuleInstall("h2o", "pip", usage="ML",
                      purpose="H2O, Fast Scalable Machine Learning, for python"),
        #
        # June 2016
        #
        ModuleInstall("kabuki", "pip", usage="ML",
                      purpose="kabuki is a python toolbox that allows easy creation of hierarchical bayesian "
                      + "models for the cognitive sciences."),
        ModuleInstall("HDDM", "wheel", mname="hddm", usage="ML",
                      purpose="HDDM is a python module that implements Hierarchical Bayesian estimation of Drift Diffusion Models."),
        ModuleInstall("reportlab", "wheel",
                      purpose="This is the ReportLab PDF Toolkit. It allows rapid creation of rich PDF documents, and also creation of "
                      + "charts in a variety of bitmap and vector formats."),
        ModuleInstall("biopython", "wheel", mname="Bio", usage="BIO",
                      purpose="The Biopython Project is an international association of developers of freely available Python tools for "
                      + "computational molecular biology."),
        ModuleInstall("grako", "wheel",
                      purpose="Grako (for grammar compiler) is a tool that takes grammars in a variation of EBNF "
                      + "as input, and outputs memoizing (Packrat) PEG parsers in Python."),
        ModuleInstall("expressions", "pip",
                      purpose="Arithmetic expression parser library. Embed customized expression evaluation "
                      + "into your application or library."),
        ModuleInstall("cubes", "pip",
                      purpose="Cubes is a light-weight Python framework and set of tools for development of reporting and "
                      + "analytical applications, "
                      + "Online Analytical Processing (OLAP), multidimensional analysis and browsing of aggregated data. "
                      + "It is part of Data Brewery."),
        ModuleInstall("MDP", "pip", usage="ML",
                      purpose="Python data processing framework."),
        ModuleInstall("pyamg", "wheel", usage="OPTIM",
                      purpose="PyAMG is a library of Algebraic Multigrid (AMG) solvers with a convenient Python interface."),
        ModuleInstall("PySAL", "pip", mname="pysal", usage="MATHS",
                      purpose="PySAL is an open source library of spatial analysis functions written in Python intended "
                      + "to support the development of high level applications."),
        #
        # July 2016
        #
        ModuleInstall("hyperopt", "github", "hyperopt", usage="ML",
                      purpose="Hyperopt is a Python library for serial and parallel optimization over awkwardsearch spaces, "
                      + "which may include real-valued, discrete, and conditional dimensions."),
        #
        # August 2016
        #
        ModuleInstall('incremental', 'pip',
                      purpose="Incremental is a small library that versions your Python projects."),
        ModuleInstall('constantly', 'pip',
                      purpose="Symbolic constants in Python"),
        ModuleInstall('PyDispatcher', 'pip', mname="pydispatcher",
                      purpose="Multi-producer-multi-consumer signal dispatching mechanism"),
        ModuleInstall('hyperlink', 'pip',
                      purpose="A featureful, correct URL for Python."),
        ModuleInstall('pyhamcrest', 'pip', purpose="for twisted"),
        ModuleInstall('Twisted', 'wheel', mname="twisted",
                      purpose="An asynchronous networking framework written in Python"),
        ModuleInstall('parsel', 'pip',
                      purpose="Parsel is a library to extract data from HTML and XML using XPath and CSS selectors."),
        ModuleInstall('pyasn1-modules', 'pip', mname="pyasn1_modules",
                      purpose="A collection of ASN.1-based protocols modules."),
        ModuleInstall('service_identity', 'pip',
                      purpose="Service identity verification for pyOpenSSL."),
        ModuleInstall("scrapy", "pip", usage="ML",
                      purpose="A high-level Web Crawling and Web Scraping framework"),

        #
        # 2015-08-11
        #
        ModuleInstall("sklearn_contrib_lightning", "wheel", mname="lightning", usage="ML",
                      purpose="large-scale linear classification, regression and ranking"),
        ModuleInstall("imbalanced-learn", "pip",
                      mname="imblearn", usage="ML",
                      purpose="imbalanced-learn is a python package offering a number of re-sampling "
                      + "techniques commonly used in datasets showing strong between-class imbalance. "
                      + "It is compatible with scikit-learn and is part of scikit-learn-contrib projects."),
        ModuleInstall("forestci", "pip", usage="ML",
                      purpose="Python module for calculating variance and adding confidence intervals "
                      + "to scikit-learn random forest regression "
                      + "or classification objects. The core functions calculate an in-bag and error bars for random forest objects"),
        #
        # 2015-08-18: weidelin.core
        #
        ModuleInstall("ZConfig", "pip",
                      purpose="Structured Configuration Library"),
        ModuleInstall("transaction", "pip",
                      purpose="Transaction management for Python. "
                      + "This package contains a generic transaction implementation for Python. It is mainly used by the ZODB."),
        ModuleInstall("zc.lockfile", "pip",
                      purpose="Basic inter-process locks"),
        ModuleInstall("zodbpickle", "wheel",
                      purpose="This package presents a uniform pickling interface for ZODB."),
        ModuleInstall("ZODB", "pip",
                      purpose="The Zope Object Database provides an object-oriented database for Python that "
                      + "provides a high-degree of transparency."),
        ModuleInstall("zdaemon", "pip",
                      purpose="Daemon process control library and tools for Unix-based systems"),
        ModuleInstall("trollius", "pip",
                      purpose="Port of the Tulip project (asyncio module, PEP 3156)"),
        ModuleInstall("ZEO", "pip",
                      purpose="ZEO provides a client-server storage implementation for ZODB."),
        ModuleInstall("ZODB3", "pip",
                      purpose="ZODB3 - Meta release for ZODB, persistent, BTrees and ZEO"),
        # ModuleInstall("wendelin.core", "wheel", source="2",
        #               purpose="Out-of-core NumPy arrays. " +
        #               "Wendelin.core allows you to work with arrays bigger than RAM and local disk. Bigarrays are persisted to storage, " +
        #               "and can be changed in transactional manner."),
        ModuleInstall("pystorm", "pip",
                      purpose="Battle-tested Apache Storm Multi-Lang implementation for Python."),
        ModuleInstall("Fabric3", "pip", mname="fabric",
                      purpose="Fabric is a simple, Pythonic tool for remote execution and deployment (py2.7/py3.4+ compatible fork)"),
        ModuleInstall("texttable", "pip",
                      purpose="module for creating simple ASCII tables"),
        ModuleInstall("streamparse", "pip",
                      purpose="Streamparse lets you run Python code against real-time streams of data via Apache Storm."),
        #
        # 2016-09-20: treeinterpreter
        #
        ModuleInstall("treeinterpreter", "pip", usage="ML",
                      purpose="Package for interpreting scikit-learn's decision tree and random forest predictions. Allows decomposing "
                      + "each prediction into bias and feature contribution components"),

        #
        # 2016-11/12
        #
        ModuleInstall("update_checker", "pip",
                      purpose="A python module that will check for package updates."),
        ModuleInstall("stopit", "pip",
                      purpose="Timeout control decorator and context managers, raise any exception in another thread."),
        ModuleInstall("TPOT", "pip", mname="tpot", usage="ML",
                      purpose="Consider TPOT your Data Science Assistant. TPOT is a Python tool that automatically "
                      + "creates and optimizes machine learning pipelines using genetic programming."),
        ModuleInstall("category_encoders", "pip", usage="ML",
                      purpose="A set of scikit-learn-style transformers for encoding categorical "
                      + "variables into numeric by means of different techniques."),
        #
        # 2017-01/02
        #
        # ModuleInstall(
        #    'backports.wearkef', 'pip', purpose="This package provides backports of new features in Python's weakref " +
        #    "module under the backports namespace. (for tensorflow)"),
        ModuleInstall('absl-py', 'pip', mname="absl",
                      purpose="Collection of Python library code for building Python applications."),
        ModuleInstall('protobuf', 'pip',
                      purpose="Protocol Buffers are Google's data interchange format"),
        ModuleInstall("gast", "pip",
                      purpose="A generic AST to represent Python2 and Python3’s Abstract Syntax Tree(AST)."),
        ModuleInstall("grpcio", "wheel",
                      purpose="Package for gRPC Python."),
        ModuleInstall("astor", "pip",
                      purpose="astor is designed to allow easy manipulation of Python source via the AST."),
        ModuleInstall("tensorboard", "pip", usage="DATA/ML",
                      purpose="TensorBoard lets you watch Tensors Flow"),
        ModuleInstall("tensorflow-estimator", "pip", usage="DEEP LEARNING",
                      mname="tensorflow_estimator", purpose="Deep Learning from Google"),
        ModuleInstall("tensorflow", "wheel", usage="DEEP LEARNING",
                      purpose="Deep Learning from Google"),
        ModuleInstall("edward", "pip", usage="DATA/ML",
                      purpose="Edward is a Python library for probabilistic modeling, inference, and criticism"),
        ModuleInstall("python_Levenshtein", "wheel", mname="Levenshtein",
                      purpose="C implementation of Levenshtein distance."),
        ModuleInstall("fuzzywuzzy", "pip",
                      purpose="Fuzzy string matching in python"),
        ModuleInstall("smmap", "pip",
                      purpose="A pure python implementation of a sliding window memory map manager"),
        ModuleInstall("gitdb", "pip",
                      purpose="GitDB is a pure-Python git object database"),
        ModuleInstall("GitPython", "pip", mname="git",
                      purpose="Python Git Library"),
        ModuleInstall("git-pandas", "pip", mname="gitpandas",
                      purpose=""),
        ModuleInstall("nbdime", "pip",
                      purpose="Comparison of notebooks"),
        ModuleInstall("cntk", "wheel", usage="DATA/ML",
                      overwrite="https://cntk.ai/PythonWheel/CPU-Only/cntk-2.0-cp{0}{1}-cp{0}{1}m-win_amd64.whl",
                      purpose="Deep Learning from Microsoft "
                      + "see https://docs.microsoft.com/en-us/cognitive-toolkit/setup-windows-python"),

        #
        # 2017-05/23
        #
        ModuleInstall('filelock', 'pip',
                      purpose="A platform independent file lock."),
        ModuleInstall('fasttext', 'wheel', usage="DATA/ML",
                      purpose="fastText is a library for efficient learning of word representations and sentence classification."),
        ModuleInstall('fastrlock', 'wheel',
                      purpose="Fast, re-entrant optimistic lock implemented in Cython (cupy)"),
        ModuleInstall('cupy', 'wheel', usage="DATA/DML",
                      purpose="CuPy : NumPy-like API accelerated with CUDA"),
        ModuleInstall('chainer', 'pip', usage="DATA/DML",
                      purpose="A flexible framework of neural networks (GPU)"),
        ModuleInstall('sfepy', 'wheel', usage="OPTIM",
                      purpose="Simple Finite Elements in Python"),
        ModuleInstall('palettable', 'pip',
                      purpose="Color palettes for Python"),
        ModuleInstall('monty', 'pip',
                      purpose="Monty is the missing complement to Python."),
        ModuleInstall('spglib', 'wheel',
                      purpose="Python bindings for C library for finding and handling crystal symmetries"),
        ModuleInstall('pymatgen', 'wheel', usage="PHYS",
                      purpose="Pymatgen (Python Materials Genomics) is a robust, open-source Python library for materials analysis."),
        ModuleInstall('gvar', 'wheel',
                      purpose="Utilities for manipulating correlated Gaussian random variables."),
        ModuleInstall('lsqfit', 'wheel', usage="OPTIM",
                      purpose="Utilities for nonlinear least-squares fits"),
        ModuleInstall('qutip', 'wheel', usage="OPTIM",
                      purpose="QuTiP is open-source software for simulating the dynamics of open quantum systems."),
        ModuleInstall('pyemd', 'wheel',
                      purpose="A Python wrapper for Ofir Pele and Michael Werman's implementation of the Earth Mover's Distance."),
        ModuleInstall('pint', 'pip',
                      purpose="Physical quantities module"),
        ModuleInstall('traits', 'pip',
                      purpose="explicitly typed attributes for Python"),
        ModuleInstall(
            'sparse', 'pip', purpose="This implements sparse arrays of arbitrary dimension on top of numpy and scipy.sparse"),
        ModuleInstall('hyperspy', 'wheel',
                      purpose="HyperSpy is an open source Python library which provides tools to facilitate "
                      + "the interactive data analysis of multi-dimensional "
                      + "datasets that can be described as multi-dimensional arrays of a given "
                      + "signal (e.g. a 2D array of spectra a.k.a spectrum image)."),
        ModuleInstall('emcee', 'pip',
                      purpose="Kick ass affine-invariant ensemble MCMC sampling"),
        ModuleInstall('h5netcdf', 'pip', purpose="netCDF4 via h5py"),
        ModuleInstall('HoloPy', 'wheel', mname="holopy",
                      purpose="Hologram processing and light scattering in python"),
        ModuleInstall('sounddevice', 'wheel',
                      purpose="This Python module provides bindings for the PortAudio library and a few "
                      + "convenience functions to play and record NumPy arrays containing audio signals."),
        ModuleInstall('drawtree', 'pip', usage="VIZ",
                      purpose="Draw binary tree in plain text"),
        ModuleInstall('JPype1', 'wheel', purpose="A Python to Java bridge."),
        ModuleInstall('pyflux', 'wheel', usage="DATA/ML",
                      purpose="An open source time series library for the Python Programming Language"),
        ModuleInstall('Rtree', 'wheel', usage="MATHS", mname="rtree",
                      purpose="R-Tree spatial index for Python GIS"),
        #
        # 2017-08/10
        #
        ModuleInstall('foolbox', 'pip', usage="ML",
                      purpose="Foolbox is a Python toolbox to create adversarial examples that fool neural networks."),
        #
        # 2017-08/25
        #
        # kealib: https://bitbucket.org/chchrsc/kealib/wiki/Home
        ModuleInstall('iso8601', 'pip',
                      purpose="Simple module to parse ISO 8601 dates"),
        ModuleInstall('translationstring', 'pip',
                      purpose="Utility library for i18n relied on by various Repoze and Pyramid packages"),
        ModuleInstall('colander', 'pip',
                      purpose="A simple schema-based serialization and deserialization library"),
        #
        ModuleInstall('climate-toolbox', 'pip', mname="climate_toolbox",
                      purpose="Command-line utilities (turn function into command line)"),
        ModuleInstall('knnimpute', 'pip', usage="ML",
                      purpose="k-Nearest Neighbor imputation"),
        # Depends on theano - not maintained anymore.
        # ModuleInstall('fancyimpute', 'pip', usage="ML",
        #               purpose="Matrix completion and feature imputation algorithms"),
        ModuleInstall('mnist', 'pip',
                      purpose="Python utilities to download and parse the MNIST dataset"),
        #
        # 2017-11
        #
        ModuleInstall('torch', 'wheel', usage="DEEP LEARNING",
                      purpose="PyTorch is a deep learning framework that puts Python first."),
        ModuleInstall('torchvision', 'pip', usage="DEEP LEARNING",
                      purpose="image and video datasets and models for torch deep learning"),
        ModuleInstall('fairtest', 'wheel', usage="ML", source="2",
                      purpose="FairTest enables developers or auditing entities to discover and test "
                      + "for unwarranted associations between an algorithm's outputs and certain user "
                      + "subpopulations identified by protected features."),
        ModuleInstall('libtiff', 'wheel',
                      purpose="PyLibTiff is a package that provides: a wrapper to the libtiff library to "
                      + "Python using ctypes, a pure Python module for reading and writing TIFF and LSM files. "
                      + "The images are read as numpy.memmap objects so that it is possible to open images "
                      + "that otherwise would not fit to computers RAM. Both TIFF strips and tiles are "
                      + "supported for low-level data storage."),
        ModuleInstall('OpenImageIO', 'wheel',
                      purpose="A library for reading and writing images with emphasis on animation and visual effects."),
        ModuleInstall('urwid', 'pip', purpose="for pyfm"),
        ModuleInstall('pyfm', 'wheel', usage="ML",
                      purpose="Factorization Machine"),
        ModuleInstall('indexed_gzip', 'wheel',
                      purpose="Fast random access of gzip files."),
        ModuleInstall(
            'mockextras', 'pip', purpose="Extensions to the mock library"),
        ModuleInstall(
            'swiglpk', 'wheel', purpose="swiglpk - Simple swig bindings for the GNU Linear Programming Kit"),
        ModuleInstall('optlang', 'pip', purpose="Formulate optimization problems using sympy expressions "
                      + "and solve them using interfaces to third-party optimization software (e.g. GLPK)."),
        ModuleInstall('depinfo', 'pip', purpose="A utility Python package intended for other library packages. "
                      + "Provides a function that when called with your package name, will print platform and dependency information."),
        ModuleInstall(
            'cobra', 'wheel', purpose="Constraint-based reconstruction and analysis in python."),
        ModuleInstall('spectrum', 'wheel', purpose="Spectrum contains tools to estimate Power Spectral Densities using "
                      + "methods based on Fourier transform, Parametric methods or eigenvalues analysis"),
        ModuleInstall('properties', 'pip',
                      purpose="An organizational aid and wrapper for validation and tab completion of class properties"),
        ModuleInstall('pymkl', 'pip', mname="pyMKL",
                      purpose="Python wrapper of Intel MKL routines"),
        ModuleInstall('pymatsolver', 'pip',
                      purpose="A (sparse) matrix solver for python."),
        ModuleInstall('vectormath', 'pip',
                      purpose="Vector math utilities for python (used by discretize)."),
        ModuleInstall('discretize', 'wheel',
                      purpose="A python package for finite volume discretization."),
        ModuleInstall('recordclass', 'wheel',
                      purpose="A mutable variant of collections.namedtuple, which supports assignments."),
        ModuleInstall('pyeda', 'wheel',
                      purpose="A library for electronic design automation."),
        ModuleInstall('Polygon3', 'wheel', mname="polygon", purpose="Handles polygonal shapes in 2D. This library is free "
                      + "for non-commercial use only."),

        #
        # 2017-12
        #
        ModuleInstall('chainercv', 'pip', usage="DATA/DML",
                      purpose="ChainerCV is a deep learning based computer vision library built on top of Chainer."),
        ModuleInstall(
            'fcn', 'pip', purpose="Fully Convolutional Networks", usage="ML"),

        #
        # 2018-14
        #
        ModuleInstall('pyltr', 'pip', usage="ML", purpose="pyltr is a Python learning-to-rank toolkit "
                      "with ranking models, evaluation metrics, data wrangling helpers, and more."),
        ModuleInstall('typing_extensions', 'pip', usage="ML",
                      purpose="Typing Extensions - Backported and Experimental Type Hints for Python"),
        ModuleInstall('onnx', 'wheel2', usage="ML",
                      purpose="Open Neural Network Exchange"),
        ModuleInstall('onnxmltools', 'pip', usage="ML",
                      purpose="Converts Machine Learning models to ONNX"),
        ModuleInstall('sklearn-onnx', 'pip', usage="ML",
                      purpose="Converts scikit-learn Machine Learning models to ONNX"),
        ModuleInstall('keras-onnx', 'pip', usage="ML",
                      purpose="Converts keras Machine Learning models to ONNX"),
        ModuleInstall('tensorflow-onnx', 'pip', usage="ML",
                      purpose="Converts tensorflow Machine Learning models to ONNX"),

        #
        # 2019-01
        #
        ModuleInstall('openTSNE', 'wheel', usage="ML", purpose="Fast t-SNE"),
        ModuleInstall('python-louvain', 'pip', usage="ML",
                      purpose="Louvain algorithm"),

        #
        # 2021-11
        #
        ModuleInstall('onnxruntime_training', 'wheel',
                      usage="ML", purpose="onnxruntime training"),
        ModuleInstall('aten_op_executor', 'wheel', usage="ML",
                      purpose="onnxruntime training"),
        ModuleInstall('torch_interop_utils', 'wheel', usage="ML",
                      purpose="onnxruntime training"),
    ]

    if sys.version_info[0] == 2:
        mod.extend([
            # (for dpark)"),
            ModuleInstall("google-common", "pip", mname="google_common",
                          purpose="Google namespace package"),
            ModuleInstall("mesos.interface", "pip",
                          purpose="Mesos interfaces (for dpark)"),
            ModuleInstall("pymesos", "pip",
                          purpose="Mesos interfaces (for dpark)"),
            # ModuleInstall("dpark", "wheel2",
            # purpose="DPark is a Python clone of Spark, MapReduce(R) alike
            # computing framework supporting iterative computation., see
            # https://github.com/douban/dpark", usage="DATA/ML"),
        ])

    if sys.platform.startswith("win"):
        mod.append(ModuleInstall("comtypes", "pip",
                                 purpose="Pure Python COM package"))
        mod.append(ModuleInstall("xlwings", "pip",
                                 purpose="reads/writes Excel files", usage="WINDOWS") if sys.platform.startswith("win") else None),
        mod.append(ModuleInstall("VideoCapture", "wheel",
                                 purpose="A Win32 Python Extension for Accessing Video Devices", usage="VIDEO"))
        mod.append(ModuleInstall("jaraco.structures", "pip",
                                 purpose="jaraco.structures"))
        mod.append(ModuleInstall("jaraco.classes", "pip",
                                 purpose="Split from another package jaraco."))

    mod.append(ModuleInstall(
        "PyJWT", "pip", mname="jwt", purpose="JSON Web Token library for Python 3."))

    return [_ for _ in mod if _ is not None]



# -*- coding: utf-8 -*-
"""
@file
@brief Defines a set of modules for more machine learning, tools, networks, visualization.
"""
import sys
from ..installhelper.module_install import ModuleInstall
from .config_helper import is_64bit


def extended_set():
    """
    list of modules to install, an rich set, to work with data and more, it requires the modules in set *small*
    """
    mod = [
        ModuleInstall("m2r", "pip",
                      purpose="M2R converts a markdown file including reStructuredText (rst) markups to a valid rst format."),
        ModuleInstall("CommonMark", "pip", mname="commonmark",
                      purpose="Python parser for the CommonMark Markdown spec"),
        ModuleInstall("recommonmark", "pip",
                      purpose="For pymc3???"),
        ModuleInstall("algopy", "pip", usage="OPTIM",
                      purpose="ALGOPY: Taylor Arithmetic Computation and Algorithmic Differentiation"),
        ModuleInstall("numdifftools", "pip", usage="OPTIM",
                      purpose="Solves automatic numerical differentiation problems in one or more variables."),
        ModuleInstall("numpydoc", "pip",
                      purpose="Sphinx extension to support docstrings in Numpy format"),
        ModuleInstall("Automat", "pip", mname="automat",
                      purpose="Self-service finite-state machines for the programmer on the go."),
        ModuleInstall("guidata", "pip" if not sys.platform.startswith("win") else "wheel2",
                      purpose="Automatic graphical user interfaces generation for easy dataset editing and display"),
        ModuleInstall(
            "guiqwt", "wheel", purpose="Efficient 2D plotting Python library based on PythonQwt (Spyder)"),
        ModuleInstall(
            "QtAwesome", "pip", mname="qtawesome",
            purpose="QtAwesome enables iconic fonts such as Font Awesome and Elusive Icons in PyQt and PySide applications."),
        ModuleInstall(
            "natgrid", "wheel", mname="mpl_toolkits.natgrid",
            purpose="Python interface to NCAR natgrid library (for matplotlib)"),
        ModuleInstall(
            "py", "pip", purpose="library with cross-python path, ini-parsing, io, code, log facilities"),
        ModuleInstall("pluggy", "pip",
                      purpose="plugin and hook calling mechanisms for python"),
        ModuleInstall("atomicwrites", "pip", purpose="Atomic file writes."),
        ModuleInstall("pytest", "pip",
                      purpose="pytest: simple powerful testing with Python"),
        ModuleInstall("parameterized", "pip",
                      purpose="parameterized: For everything. Parameterized testing "
                      "for nose, parameterized testing for py.test, parameterized testing for unittest."),
        ModuleInstall(
            "blist", "wheel",
            purpose="a list-like type with better asymptotic performance and similar performance on small lists"),
        ModuleInstall(
            "blz", "wheel",
            purpose="blz: a compressed data container"),
        ModuleInstall("pamela", "pip",
                      purpose="An interface to the Pluggable Authentication Modules (PAM) " +
                      "library on linux, written in pure python (using ctypes)")
        if not sys.platform.startswith("win") else None,
        ModuleInstall("async_generator", "pip", purpose="for jupyterhup"),
        ModuleInstall("python_oauth2", "pip", purpose="for jupyterhup"),
        ModuleInstall(
            "jupyterhub", "pip", purpose="JupyterHub: A multi-user server for Jupyter notebooks", usage="JUPYTER")
        if not sys.platform.startswith("win") else None,
        ModuleInstall('rpy2', 'wheel', purpose="interact with R (R_HOME needs to be set up on Linux)",
                      usage="DATA/ML"),
        ModuleInstall('python-pptx', 'pip', mname="pptx",
                      purpose="read/write PowerPoint presentation"),
        ModuleInstall(
            'python-docx', 'pip', mname="docx", purpose="read/write Word document"),
        # ModuleInstall('flasksphinx', 'pip', purpose="serves Sphinx
        # documentation through a Flask server"), # issue with Python 3
        ModuleInstall(
            'cffi', 'wheel', usage="C++", purpose="Foreign Function Interface for Python calling C code."),
        ModuleInstall(
            'datashape', 'pip', purpose="A data description language."),
        ModuleInstall(
            'ordereddict', 'pip', purpose="Python's collections.OrderedDict") if sys.version_info[0] == 2 else None,
        ModuleInstall(
            'cyordereddict', 'wheel', purpose="Cython implementation of Python's collections.OrderedDict"),
        ModuleInstall('dynd', 'wheel',
                      purpose="DyND-Python, a component of the Blaze project, " +
                      "is the Python exposure of the DyND dynamic multi-dimensional array library.")
        if sys.version_info[0] >= 3 else None,
        ModuleInstall("mpmath", "pip",
                      purpose="mpmath is a free (BSD licensed) Python library for real and complex " +
                      "floating-point arithmetic with arbitrary precision."),
        ModuleInstall(
            'sympy', 'pip', purpose="SymPy is a Python library for symbolic mathematics."),
        ModuleInstall('gmpy2', 'wheel',
                      purpose="big real numbers (issue on Linux and Anaconda)"),
        ModuleInstall('llvmlite', 'wheel',
                      purpose="lightweight wrapper around basic LLVM functionality, check issue " +
                      "https://github.com/cmderdev/cmder/issues/490 for missing api-ms-win-crt-runtime-l1-1-0.dll"),
        ModuleInstall('numba', 'wheel', usage="C++",
                      purpose="Numba is an Open Source NumPy-aware optimizing compiler for Python sponsored by Continuum Analytics, Inc."),
        ModuleInstall('scikit-image', 'wheel', mname='skimage',
                      purpose="scikit-image is a collection of algorithms for image processing."),
        ModuleInstall(
            'cvxopt', 'wheel', purpose="linear, quadratique optimization", usage="OPTIM"),
        ModuleInstall(
            'PyWavelets', 'wheel', mname='pywt', purpose="wavelets computation", usage="DATA/ML"),
        ModuleInstall('pyclustering', 'pip',
                      purpose="many kinds of clustering (Optics, DBScan, x-means, ...)", usage="DATA/ML"),
        ModuleInstall(
            'pycosat', 'wheel', purpose="PicoSAT is a popular SAT solver written by Armin Biere in pure C."),
        ModuleInstall('pyshp', 'pip', mname='shapefile',
                      purpose="Pure Python read/write support for ESRI Shapefile format"),
        ModuleInstall('descartes', 'pip',
                      purpose="Use Shapely or GeoJSON-like geometric objects as matplotlib paths and patches"),
        ModuleInstall('geopandas', 'pip',
                      purpose="GeoPandas is an open source project to make working with geospatial data in python easier. "),
        ModuleInstall(
            'vispy', 'pip', purpose="Vispy is a high-performance interactive 2D/3D data visualization library."),
        ModuleInstall(
            'selenium', 'pip', purpose="Python wrapper for Selenium", usage="NETWORK"),
        ModuleInstall(
            'splinter', 'pip', purpose="browser abstraction for web acceptance testing", usage="NETWORK"),
        ModuleInstall(
            'pygame', 'wheel', purpose="GUI, interface for games (needs to be installed from www.pygame.org on Linux)", usage="GUI"),
        ModuleInstall(
            'Kivy', 'wheel', mname='kivy', usage="GUI",
            purpose="GUI, interface for games, mobile (use sudo apt-get install python3-kivy on Linux)"),
        ModuleInstall('kivy-garden', 'pip', mname='kivy.garden',
                      purpose="Garden tool for kivy flowers.", usage="GUI"),
        ModuleInstall(
            'py4j', 'pip', purpose="Enables Python programs to dynamically access arbitrary Java objects"),
        ModuleInstall(
            'lockfile', 'pip', purpose="Platform-independent file locking module"),
        ModuleInstall('python-daemon', 'pip', mname="daemon",
                      purpose="Library to implement a well-behaved Unix daemon process (for luigi)"),
        ModuleInstall('cached_property', 'pip',
                      purpose="A decorator for caching properties in classes (for luigi)"),
        ModuleInstall('luigi', 'pip',
                      purpose="workflows, data workflows", usage="WORKFLOW"),
        #
        ModuleInstall('setproctitle', 'wheel',
                      purpose="A Python module to customize the process title"),
        # thrift only works only for Python 2.7
        ModuleInstall(
            'ply', 'pip', purpose="Python Lex & Yacc (for thrifty)"),
        ModuleInstall(
            'thrift', 'pip', purpose="Python bindings for the Apache Thrift RPC system"),
        ModuleInstall(
            'thriftpy', 'pip', purpose="pure python implemention of Apache Thrift."),
        # ModuleInstall('airflow', 'pip'),  # does not work on Python 3
        ModuleInstall(
            'branca', 'pip', purpose="", usage="VIZ"),
        ModuleInstall('folium', 'pip', usage="VIZ",
                      purpose="This library is a spinoff from folium, that would host the " +
                      "non-map-specific features. It may become a HTML+JS generation " +
                      "library in the future."),
        ModuleInstall(
            'osmapi', 'pip', purpose="Python wrapper for the OSM API", usage="VIZ"),
        ModuleInstall('geographiclib', 'pip',
                      purpose="This implements algorithms for geodesics (Karney, 2013) for solving " +
                      "the direct and inverse problems for an ellipsoid of revolution."),
        ModuleInstall('geopy', 'pip',
                      purpose="Python Geocoding Toolbox", usage="VIZ"),
        ModuleInstall('geojson', 'pip',
                      purpose="Functions for encoding and decoding GeoJSON formatted data"),
        # Deprecated
        # ModuleInstall('basemap', 'wheel', mname='mpl_toolkits.basemap',
        #               purpose="maps extension for matplotlib", usage="VIZ"),
        ModuleInstall("python3-linkedin", "pip", mname="linkedin",
                      purpose="Python Interface to the LinkedIn API"),
        # access to linkedin
        ModuleInstall(
            "oauthlib", "pip", purpose="A generic, spec-compliant, thorough implementation of the OAuth request-signing logic"),
        ModuleInstall("requests-oauthlib", "pip", mname="requests_oauthlib",
                      purpose="OAuthlib authentication support for Requests."),
        ModuleInstall("antlr4-python3-runtime", "pip",
                      mname="antlr4", purpose="antlr4 runtime, grammar parser"),
        # ModuleInstall("unqlite",                    "pip"),   #
        # key/value store (NoSQL)
        ModuleInstall("pycontracts", "pip", mname="contracts",  # version="1.7.6",
                      purpose="PyContracts is a Python package that allows to declare constraints on function parameters " +
                      "and return values, setup for version 1.7.7 is bugged"),
        #
        ModuleInstall(
            "ecdsa", "pip", purpose="ECDSA cryptographic signature library (pure python)"),
        ModuleInstall("winrandom", "wheel",
                      purpose="This module gives direct access to Windows Cryptographic API CryptGetRandom() function, " +
                      "which is cryptographically strong pseudo-random number generator (PRNG) on Windows:"),
        ModuleInstall("pycrypto", "wheel2", mname="Crypto",
                      purpose="Cryptographic modules for Python (not available on x64 and Python 3)"),
        ModuleInstall("pycryptodomex", "pip", mname="Cryptodome",
                      purpose="Cryptographic modules for Python (not available on x64 and Python 3)"),
        ModuleInstall("xxhash", "wheel",
                      purpose="xxHash is an Extremely fast Hash algorithm, running at RAM speed limits."),
        ModuleInstall("cryptography", "pip",
                      purpose="cryptography is a package which provides cryptographic recipes and primitives to Python developers."),
        ModuleInstall(
            "pyasn1", "pip", purpose="ASN.1 types and codecs (for pysnmp)"),
        ModuleInstall("asn1crypto", "pip",
                      purpose="Fast ASN.1 parser and serializer with definitions for private keys, " +
                      "public keys, certificates, CRL, OCSP, CMS, PKCS#3, PKCS#7, PKCS#8, PKCS#12, PKCS#5, X.509 and TSP"),
        ModuleInstall("PyNaCl", "pip", mname="nacl",
                      purpose="Python binding to the Networking and Cryptography (NaCl) library", usage="NETWORK"),
        ModuleInstall("bcrypt", "pip",
                      purpose="Modern password hashing for your software and your servers"),
        ModuleInstall("paramiko", "pip",
                      purpose="SSH2 protocol library", usage="NETWORK"),
        #
        #
        # 2015-02-05
        #
        ModuleInstall("autopy3", "wheel",
                      purpose="A simple, cross-platform GUI automation toolkit for Python 3 " +
                      "(issue on Linux and Anaconda)") if sys.version_info[0] >= 3 else None,  # simulate events
        # large double
        ModuleInstall("bigfloat", "wheel",
                      purpose="big float (issue with Linux and Anaconda)"),
        # convex optimization, depends on CVXOPT
        ModuleInstall(
            "scs", "wheel", purpose="Solves convex cone programs via operator splitting.", usage="OPTIM"),
        ModuleInstall(
            "ecos", "wheel", purpose="ECOS is a numerical software for solving convex second-order cone programs (SOCPs)", usage="OPTIM"),
        ModuleInstall(
            "CVXcanon", "wheel", purpose="A low-level library to perform the matrix building step in cvxpy, " +
            "a convex optimization modeling software.", usage="OPTIM") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall(
            "fastcache", "wheel", purpose="C implementation of Python 3 lru_cache for Python."),
        ModuleInstall(
            "multiprocess", "wheel", purpose="better multiprocessing and multithreading in python"),
        ModuleInstall("osqp", "wheel",
                      purpose="The OSQP (Operator Splitting Quadratic Program) solver is a numerical optimization package for solving " +
                      "quadratic problems") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall("cvxpy", "pip", usage="OPTIM",
                      purpose="linear, quadratic optimization, depends on cvxopt") if sys.version_info[:2] >= (3, 5) else None,
        # to install packages with conda
        ModuleInstall("libLAS", "wheel", mname="liblas",
                      purpose="libLAS is a C/C++ library for reading and writing the very common LAS LiDAR format."),
        ModuleInstall(
            "liblinear", "wheel", purpose="A Library for Large Linear Classification"),
        ModuleInstall("marisa_trie", "wheel",
                      purpose="Static memory-efficient & fast Trie-like structures for Python (based on marisa-trie C++ library)"),
        ModuleInstall(
            "mlpy", "wheel", purpose="mlpy is a Python module for Machine Learning built on top of NumPy/SciPy, has wavelets"),
        ModuleInstall(
            "pygit2", "wheel", purpose="Pygit2 is a set of Python bindings to the libgit2 shared library, " +
            "libgit2 implements the core of Git."),
        ModuleInstall(
            "pymongo", "wheel", purpose="Python wrapper for MongoDB", usage="NoSQL"),
        ModuleInstall("psycopg2", "wheel",
                      purpose="Python-PostgreSQL Database Adapter"),
        ModuleInstall(
            "PyOpenGL", "wheel", mname="OpenGL", purpose="use OpenGL in Python"),
        ModuleInstall(
            "PyOpenGL_accelerate", "wheel", mname="OpenGL_accelerate", purpose="Acceleration code for PyOpenGL"),
        ModuleInstall('pymc', 'wheel', web="https://github.com/pymc-devs/pymc",
                      purpose="Monte Carlo computation", usage="DATA/ML") if sys.version_info[0] >= 3 else None,
        ModuleInstall('autograd', 'pip',
                      purpose="Efficiently computes derivatives of numpy code."),
        # The following package rely on theano (deprecated).
        # ModuleInstall('pymc3', 'github', "pymc-devs", web="https://github.com/pymc-devs/pymc3",
        #              purpose="Monte Carlo computation (Python 3 only)", usage="DATA/ML") if sys.version_info[0] >= 3 else None,
        # ModuleInstall('pysterior', 'pip',
        #              purpose="pysterior is a machine learning library for Python which aims to make Bayesian parametric regression and " +
        #               "classification models accessible and easy to use. The library allows users to construct " +
        #               "supervised learning models using an intuitive interface similar to that used by scikit-learn.",
        #               usage="DATA/ML") if sys.version_info[0] >= 3 else None,
        ModuleInstall(
            "pyqtgraph", "pip", purpose="Scientific Graphics and GUI Library for Python, depends on PySide", usage="GUI"),
        ModuleInstall("deap", "pip", purpose="deep learning"),
        # for gensim and distributed
        ModuleInstall("jmespath", "pip", purpose="JSON Matching Expressions"),
        # for gensim
        ModuleInstall("bz2file", "pip", purpose="process bz2 files"),
        # for gensim
        ModuleInstall("smart_open", "pip",
                      purpose="Utils for streaming large files (S3, HDFS, gzip, bz2...), provides the same API for many format"),
        ModuleInstall("httpretty", "pip",
                      purpose="HTTP client mock for Python"),
        ModuleInstall("gensim", "wheel", purpose="genetic algorithm"),
        # ModuleInstall("pybrain", "pip"),   # some issues with the code
        # (relative import are not well handled in version 0.3.3
        ModuleInstall("h5py", "wheel", usage="DATA/ML",
                      purpose="The h5py package is a Pythonic interface to the HDF5 binary data format. " +
                      "Trillion-Particle Simulation."),
        ModuleInstall("Keras-Applications", "pip", mname="keras_applications", usage="DATA/ML",
                      purpose="Keras Applications is the applications module of the Keras deep learning " +
                      "library. It provides model definitions and pre-trained weights for a number of " +
                      "popular archictures, such as VGG16, ResNet50, Xception, MobileNet, and more."),
        ModuleInstall("Keras-Preprocessing", "pip", mname="keras_processing", usage="DATA/ML",
                      purpose="Keras Preprocessing is the data preprocessing and data augmentation module " +
                      "of the Keras deep learning library. It provides utilities for working with image " +
                      "data, text data, and sequence data."),
        ModuleInstall("keras", "pip", purpose="deep learning",
                      usage="DATA/ML"),
        ModuleInstall("keras-vis", "pip", mname="vis", usage="DATA/ML",
                      purpose="keras-vis is a high-level toolkit for visualizing and debugging your trained keras neural net models."),
        # Bayesian
        ModuleInstall(
            "bayespy", "pip", purpose="bayesian modelling and computation", usage="DATA/ML"),
        ModuleInstall(
            "numexpr", "wheel", purpose="Fast numerical array expression evaluator for Python, NumPy, PyTables, pandas and more."),
        #
        ModuleInstall("glue-core", "pip", mname="glue.core",
                      purpose="Multidimensional data visualzation across files"),
        ModuleInstall("glue-vispy-viewers", "pip", mname="glue_vispy_viewers",
                      purpose="Vispy-based viewers for Glue"),
        ModuleInstall("glueviz", "pip", mname="glue",
                      purpose="ploting, Multidimensional data visualzation across files", usage="DATA/ML"),
        #
        ModuleInstall("pandas-highcharts", "pip", mname="pandas_highcharts",
                      purpose="plotting in javascript and pandas", usage="VIZ"),
        #
        ModuleInstall(
            "heapdict", "pip",
            purpose="a heap with decrease-key and increase-key operations"),
        ModuleInstall(
            "chest", "pip",
            purpose="Simple on-disk dictionary"),
        ModuleInstall(
            "locket", "pip",
            purpose="File-based locks for Python for Linux and Windows"),
        ModuleInstall(
            "partd", "pip",
            purpose="Appendable key-value storage"),
        ModuleInstall(
            "dill", "pip", purpose="serialize all of python (almost), Dill extends python's pickle module for serializing " +
            "and de-serializing python objects to the majority of the built-in python types."),  # for dask
        ModuleInstall("cloudpickle", "pip",
                      purpose="Extended pickling support for Python objects") if sys.version_info[:2] >= (3, 5) else None,
        # parallel computation
        ModuleInstall(
            "dask", "pip", purpose="parallization of operations with dataframe", usage="DATA/ML"),
        ModuleInstall(
            "scoop", "pip", purpose="SCOOP (Scalable COncurrent Operations in Python) " +
            "is a distributed task module allowing concurrent parallel programming on various environments, " +
            "from heterogeneous grids to supercomputers", usage="DATA/ML"),
        #
        ModuleInstall(
            "docopt", "pip", purpose="Pythonic argument parser, that will make you smile"),
        ModuleInstall("pycurl", "wheel",
                      purpose="PycURL, a interface to the libcurl library. (for grab)"),
        ModuleInstall("markdown2", "pip", purpose="markdown parser"),
        ModuleInstall(
            "structures", "pip", purpose="User-friendly library for creating data structures."),
        ModuleInstall(
            "tzlocal", "pip", purpose="tzinfo object for the local timezone"),
        ModuleInstall(
            "funcsigs", "pip", purpose="Python function signatures from PEP362"),
        ModuleInstall(
            "apscheduler", "pip", purpose="to schedule the execution of jobs, tasks"),
        #
        # ModuleInstall("pdfminer", "pip"),  # PDF extraction (no python 3 version)
        # ModuleInstall("minecart", "pip"),  # PDF extraction (no python 3 version)
        #
        # ModuleInstall("pygauss", "pip"),  # molecule, bio-informatic,
        # requires PIL which is deprecated
        #
        # July 2015
        #
        ModuleInstall(
            "pyexecjs", "pip", mname="execjs", purpose="Run JavaScript code from Python (for pyreact)", usage="NETWORK"),
        ModuleInstall("pyreact", "pip", mname="react",
                      purpose="Python bridge to JSX & the React JavaScript library. (for pyxley)", usage="NETWORK"),
        ModuleInstall(
            "pyxley", "pip", purpose="A pure-Python SNMPv1/v2c/v3 library", usage="NETWORK"),

        #
        # 2015-08
        #
        ModuleInstall(
            "pysmi", "pip", purpose="SNMP SMI/MIB Parser (for pysnmp)"),
        ModuleInstall(
            "pysnmp", "pip", purpose="A pure-Python SNMPv1/v2c/v3 library", usage="NETWORK"),
        # pyinstaller does not install properly on Windows
        # ModuleInstall(
        #    "pyinstaller", "pip", purpose="Converts (packages) Python programs into stand-alone
        #    executables, under Windows, Linux, Mac OS X, AIX and Solaris."),
        ModuleInstall(
            "imageio", "pip", purpose="Library for reading and writing a wide range of image, video, " +
            "scientific, and volumetric data formats (for moviepy)", usage="VIDEO"),
        ModuleInstall(
            "cairocffi", "wheel2", purpose="cairocffi is a CFFI-based drop-in replacement for Pycairo, " +
            "a set of Python bindings and object-oriented API for cairo"),
        ModuleInstall("tinycss2", "pip", purpose="tinycss2 is a complete yet simple CSS parser for Python. " +
                      "It supports the full syntax and error handling for CSS 2.1 as well as some CSS 3 modules"),
        ModuleInstall("cssselect2", "pip",
                      purpose="CSS selectors for Python ElementTree"),
        ModuleInstall(
            "cairosvg", "pip", purpose="Convert your SVG files to PDF and PNG."),
        ModuleInstall(
            "gizeh", "pip", purpose="Simple Vector Graphics for Python"),
        ModuleInstall("imageio-ffmpeg", "pip", mname="imageio_ffmpeg",
                      purpose="ffmpeg", usage="VIDEO"),
        ModuleInstall(
            "proglog", "pip", purpose="Progress bar."),
        ModuleInstall(
            "moviepy", "pip", purpose="Video editing with Python", usage="VIDEO"),
        ModuleInstall(
            "xgboost", "wheel", purpose="Parallelized Stochastic Gradient Descent (only available on " +
            "Python 3 and x64)", usage="DATA/ML") if sys.version_info[0] >= 3 and is_64bit() else None,
        ModuleInstall(
            "catboost", "pip", purpose="CatBoost is a machine learning method based on gradient boosting " +
            "over decision trees.", usage="DATA/ML") if sys.version_info[0] >= 3 and is_64bit() else None,
        ModuleInstall(
            "lightgbm", "pip", purpose="Parallelized Stochastic Gradient Descent (only available on " +
            "Python 3 and x64)", usage="DATA/ML") if sys.version_info[0] >= 3 and is_64bit() else None,
        ModuleInstall("pygling", "pip",
                      purpose="to build makefile with python") if sys.version_info[0] == 2 else None,
        ModuleInstall("cuda4py", "pip",
                      purpose="Python cffi CUDA bindings and helper classes"),
        ModuleInstall("whoosh", "pip", purpose="search engine in Python"),
        ModuleInstall("pymatbridge", "pip",
                      purpose="pymatbridge is a set of python and matlab functions to allow these two systems to talk to each other"),
        ModuleInstall("scilab2py", "pip",
                      purpose="Python to Scilab bridge", usage="DATA/ML"),
        # ModuleInstall("scilab_kernel", "pip",
        #              purpose="A Scilab kernel for IPython", usage="JUPYTER"),
        # does not work
        ModuleInstall("pymssql", "wheel", usage="SQL",
                      purpose="A simple database interface for Python that builds on top of FreeTDS " +
                      "to provide a Python DB-API (PEP-249) interface to Microsoft SQL Server."),

        ModuleInstall("PyMySQL", "pip", mname="pymysql",
                      purpose="Pure-Python MySQL Driver", usage="SQL"),
        ModuleInstall("mysqlclient", "wheel", mname="MySQLdb",
                      purpose="MySQL driver written in Python which does not depend on MySQL C client libraries and " +
                      "implements the DB API v2.0 specification (PEP-249).", usage="SQL"),
        ModuleInstall("memory-profiler", "pip", mname="memory_profiler",
                      purpose="A module for monitoring memory usage of a python program", usage="PROFILING"),
        ModuleInstall("pyinstrument_cext", "pip", usage="PROFILING",
                      purpose="Pyinstrument's C extensions - reducing the overhead of statistical profilers"),
        ModuleInstall("pyinstrument", "pip", usage="PROFILING",
                      purpose="A Python profiler that records the call stack of the executing code, " +
                      "instead of just the final function in it."),
        ModuleInstall("gprof2dot", "pip", usage="PROFILING",
                      purpose="This is a Python script to convert the output from many profilers into a dot graph."),
        ModuleInstall("vprof", "pip", usage="PROFILING",
                      purpose="vprof is a Python package providing rich and interactive visualizations for various Python " +
                      "program characteristics such as running time and memory usage."),
        ModuleInstall("snakeviz", "pip",
                      purpose="SnakeViz is a browser based graphical viewer for the output of Python’s cProfile module.", usage="PROFILING"),
        ModuleInstall("httplib2", "pip",
                      purpose="A comprehensive HTTP client library."),
        ModuleInstall("rsa", "pip",
                      purpose="Pure-Python RSA implementation"),
        ModuleInstall("oauth2client", "pip",
                      purpose="The oauth2client is a client library for OAuth 2.0."),
        ModuleInstall("uritemplate", "pip",
                      purpose="URI templates"),
        ModuleInstall("jeepney", "pip",
                      purpose="This is a low-level, pure Python DBus protocol client. It has an I/O-free core, and " + \
                      "integration modules for different event loops."),
        ModuleInstall("secretstorage", "pip",
                      purpose="This module provides a way for securely storing passwords and other secrets."),
        ModuleInstall("keyring", "pip",
                      purpose="Store and access your passwords safely."),
        # ModuleInstall("keyrings.alt", "pip", mname="keyrings.alt",
        #               purpose="Alternate keyring implementations"),
        ModuleInstall("pyotp", "pip",
                      purpose="PyOTP is a Python library for generating and verifying one-time passwords."),
        ModuleInstall("param", "pip",
                      purpose="Declarative Python programming using Parameters."),
        ModuleInstall("pyviz_comms", "pip", usage="VIZ",
                      purpose="Bidirectional communication for PyViz."),
        ModuleInstall("holoviews", "pip", usage="VIZ",
                      purpose="Composable, declarative data structures for building complex visualizations easily."),
        ModuleInstall("geoviews", "pip", usage="VIZ",
                      purpose="Composable, declarative data structures for building complex visualizations easily."),
        ModuleInstall("retrying", "pip", purpose="Retrying is an Apache 2.0 licensed general-purpose retrying library, " +
                      "written in Python, to simplify the task of adding retry behavior to just about anything."),
        ModuleInstall("plotly", "pip", usage="VIZ",
                      purpose="Plotly's Python graphing library makes interactive, publication-quality graphs online. " +
                      "Examples of how to make line plots, " +
                      "scatter plots, area charts, bar charts, error bars, box plots, histograms, heatmaps, subplots, multiple-axes, " +
                      "polar charts and bubble charts."),
        ModuleInstall("colorlover", "pip", usage="VIZ",
                      purpose="Color scales for IPython notebook"),
        ModuleInstall("TA_Lib", "wheel", mname="talib",
                      purpose="This is a Python wrapper for TA-LIB based on Cython instead of SWIG."),
        ModuleInstall("cufflinks", "pip", usage="VIZ",
                      purpose="Productivity Tools for Plotly + Pandas"),
        ModuleInstall("lightning-python", "pip", mname="lightning", usage="VIZ",
                      purpose="Python client for the lightning API"),
        ModuleInstall("passlib", "pip",
                      purpose="comprehensive password hashing framework supporting over 30 schemes"),
        ModuleInstall("plac", "pip",
                      purpose="The smartest command line arguments parser in the world"),
        ModuleInstall("pyOpenSSL", "pip", mname="OpenSSL",
                      purpose="Python wrapper module around the OpenSSL library"),
        ModuleInstall("w3lib", "pip",
                      purpose="Library of web-related functions"),
        # ModuleInstall('python-cloudfiles-hubic', 'github', "Gu1", mname="cloudfiles",
        #                   web="https://github.com/Gu1/python-cloudfiles-hubic",
        #              purpose="access to Hubic"),
        # ModuleInstall('onedrive-sdk-python', 'github', "OneDrive", mname="onedrivesdk",
        #               web="https://github.com/Gu1/python-cloudfiles-hubic/",
        #               purpose="access to OneDrive"),
        # ModuleInstall('rlpy', 'pip', usage="DATA/ML",
        # purpose="RLPy is a framework to conduct sequential decision making
        # experiments. The current focus of this project lies on
        # value-function-based reinforcement learning, specifically using
        # linear function approximators (only Python 2.7)."),
        ModuleInstall('wordcloud', 'wheel', usage="VIZ",
                      purpose="A little word cloud generator in Python."),
        ModuleInstall('pytagcloud', 'pip',
                      purpose="Create beautiful tag clouds as images or HTML"),

        # distributed
        ModuleInstall('tblib', 'pip',
                      purpose="Traceback fiddling library. For now allows you to pickle tracebacks and raise exceptions with pickled " +
                      "tracebacks in different processes. This allows better error handling when running code over " +
                      "multiple processes (imagine multiprocessing, billiard, futures, celery etc)"),
        ModuleInstall("zict", "pip",
                      purpose="The dictionary / mutable mapping interface is powerful and multi-faceted."),

        # pdf
        ModuleInstall("pyPdf", "github", "sdpython", usage="PDF", pipgit=True,
                      branch="trunk", purpose="read PDF"),
        ModuleInstall("pdfrw", "pip", usage="PDF",
                      purpose="PDF file reader/writer library"),
        # 2016-05
        ModuleInstall("pydub", "pip", usage="MUSIC",
                      purpose="Pydub lets you do stuff to audio in a way that isn't stupid."),
        ModuleInstall("cobble", "pip", purpose="Cobble is a Python library that allows easy creation of data objects, " +
                      "including implementations of common methods such as __eq__ and __repr__."),
        ModuleInstall("parsimonious", "pip",
                      purpose="(Soon to be) the fastest pure-Python PEG parser I could muster"),
        ModuleInstall(
            "mammoth", "pip", purpose="Convert Word documents from docx to simple and clean HTML and Markdown"),
        #
        # 2016-06
        #
        # ModuleInstall("ipython-sql", "pip", purpose="RDBMS access via IPython", usgae="JUPYTER"),
        ModuleInstall("julia", "pip",
                      purpose="Julia/Python bridge with IPython support", usage="DATA/ML"),
        ModuleInstall("octave_kernel", "pip",
                      purpose="A Jupyter kernel for Octave."),
        ModuleInstall("oct2py", "pip",
                      purpose="Python to GNU Octave bridge --> run m-files from python.", usage="DATA/ML"),
        ModuleInstall("pg8000", "pip",
                      purpose="A Pure-Python PostgreSQL", usage="SQL"),
        ModuleInstall("PyMeta3", "pip", mname="pymeta",
                      purpose="Pattern-matching language based on OMeta for Python 3 and 2"),
        # ModuleInstall("ViTables", "pip", mname="vitables",
        #               purpose="A viewer for PyTables package"),
        ModuleInstall("pybars3", "pip", mname="pybar",
                      purpose="Handlebars.js templating"),
        ModuleInstall("db.py", "pip", mname="db.tables",
                      purpose="db.py is an easier way to interact with your databases. It makes it easier " +
                      "to explore tables, columns, views, etc. " +
                      "It puts the emphasis on user interaction, information display, and providing easy to use helper functions."),
        ModuleInstall("clyent", "pip",
                      purpose="Command line client Library for windows and posix"),
        ModuleInstall("chalmers", "pip",
                      purpose="Chalmers is an application that allows its users to monitor and control a number of processes on any " +
                      "operating system (Posix and Win32 included)"),
        ModuleInstall("pyct", "pip", purpose="A utility package"),
        ModuleInstall("colorcet", "pip",
                      purpose="colorcet is a collection of perceptually uniform colormaps for use with Python plotting programs " +
                      "like bokeh, matplotlib, holoviews, and datashader."),
        ModuleInstall("datashader", "pip", usage="VIZ",
                      purpose="Datashader is a graphics pipeline system for creating meaningful representations " +
                      "of large amounts of data.") if sys.version_info[:2] >= (3, 5) else None,
        ModuleInstall("dnspython", "pip", usage="WEB",
                      purpose="dnspython is a DNS toolkit for Python. It supports almost all record types. It can be used for queries, " +
                      "zone transfers, and dynamic updates. It supports TSIG authenticated messages and EDNS0."),
        ModuleInstall("grin", "pip", usage="CLI",
                      purpose="A grep program configured the way I like it."),
        ModuleInstall("ldap3", "pip", usage="WEB",
                      purpose="ldap3 is a strictly RFC 4510 conforming LDAP V3 pure Python client library."),
        ModuleInstall("mpi4py", "wheel",
                      purpose="MPI for Python"),
        ModuleInstall("mss", "pip", mname="mms",
                      purpose="An ultra fast cross-platform multiple screenshots module in pure python using ctypes"),
        #
        # June 2016
        #
        ModuleInstall("pyglet", "pip", usage="GUI",
                      purpose="a cross-platform windowing and multimedia library for Python"),
        ModuleInstall("geoplotlib", "pip", usage="VIZ",
                      purpose="geoplotlib is a python toolbox for visualizing geographical data and making maps"),
        ModuleInstall("leather", "pip", usage="VIZ",
                      purpose="Leather is the Python charting library for those who need charts now and don’t care if they’re perfect."),
        ModuleInstall("pythreejs", "pip", usage="VIZ",
                      purpose="A Python / ThreeJS bridge utilizing the Jupyter widget infrastructure."),
        ModuleInstall("missingno", "pip", usage="VIZ",
                      purpose="Messy datasets? Missing values? missingno provides a small toolset of flexible and easy-to-use " +
                      "missing data visualizations and utilities that allows you to get a quick visual summary of the completeness " +
                      "(or lack thereof) of your dataset."),
        ModuleInstall("vega", "pip", usage="VIZ",
                      purpose="Python/Jupyter notebook module for Vega, and Vega-Lite, Polestar, and Voyager. Notebooks " +
                      "with embedded visualizations can be viewed on github and nbviewer."),
        ModuleInstall("pydy", "pip",
                      purpose="Multibody Dynamics with Python"),
        ModuleInstall("apache-libcloud", "pip", mname="libcloud",
                      purpose="A standard Python library that abstracts away differences among multiple cloud provider APIs."),
        ModuleInstall("click-plugins", "pip", mname="click_plugins",
                      purpose="An extension module for click to enable registering CLI commands via setuptools entry-points."),
        ModuleInstall("munch", "pip",
                      purpose="A dot-accessible dictionary (a la JavaScript objects)."),
        ModuleInstall("cligj", "pip",
                      purpose="Click params for commmand line interfaces to GeoJSON"),
        ModuleInstall("brythonmagic", "pip",
                      purpose="Magics to use brython in Jupyter notebook."),
        ModuleInstall(
            "asteval", "pip", purpose="Safe, minimalistic evaluator of python expression using ast module"),
        ModuleInstall(
            "uncertainties", "pip", purpose="uncertainties allows calculations such as (2 +/- 0.1)*2 = 4 " +
            "+/- 0.2 to be performed transparently."),
        ModuleInstall(
            "lmfit", "pip", purpose="Least-Squares Minimization with Bounds and Constraints", usage="OPTIM"),
        #
        # August, September, October, December 2016
        #
        ModuleInstall('QInfer', 'pip', mname="qinfer", usage="DATA/ML",
                      purpose="QInfer is a library using Bayesian sequential Monte Carlo for quantum parameter estimation."),
        ModuleInstall('pscript', 'pip', purpose="for flexx"),
        ModuleInstall('webruntime', 'pip', purpose="for flexx"),
        ModuleInstall('dialite', 'pip', purpose="for flexx"),
        ModuleInstall('flexx', 'pip', usage="GUI",
                      purpose="Flexx is a pure Python toolkit for creating graphical user interfaces (GUI's), " +
                      "that uses web technology for its rendering. Apps are written purely in Python; Flexx' " +
                      "transpiler generates the necessary JavaScript on the fly."),
        ModuleInstall('pypng', 'pip',
                      purpose="Pure Python PNG image encoder/decoder"),
        ModuleInstall('colormath', 'pip',
                      purpose="Color math and conversion library."),
        ModuleInstall('arrow', 'pip',
                      purpose="Better dates and times for Python"),
        ModuleInstall('custom_inherit', 'pip',
                      purpose="A Python package that provides tools for inheriting docstrings in customizable ways."),
        ModuleInstall('toyplot', 'pip', usage="VIZ",
                      purpose="The kid-sized plotting toolkit for Python with grownup-sized goals."),
        ModuleInstall('images2gif', 'pip',
                      purpose="Create a GIF from a list of images."),
        ModuleInstall('hypothesis', 'pip',
                      purpose="Hypothesis is an advanced testing library for Python. It lets you write tests which are parametrized " +
                      "by a source of examples, and then generates simple and comprehensible examples that make your tests fail. " +
                      "This lets you find more bugs in your code with less work."),
        ModuleInstall('monotonic', 'pip',
                      purpose="This module provides a monotonic() function which returns the value (in fractional seconds) " +
                      "of a clock which never goes backwards. " +
                      "It is compatible with Python 2 and Python 3."),
        ModuleInstall('fasteners', 'pip',
                      purpose="A python package that provides useful locks."),
        ModuleInstall('asciitree', 'pip',
                      purpose="Draws ASCII trees."),
        ModuleInstall('numcodecs', 'wheel',
                      purpose="Numcodecs is a Python package providing buffer compression and transformation codecs for use " +
                      "in data storage and communication applications."),
        ModuleInstall('zarr', 'pip',
                      purpose="A minimal implementation of chunked, compressed, N-dimensional arrays for Python."),
        ModuleInstall('rx', 'pip',
                      purpose="Reactive Extensions (Rx) for Python (LINQ)"),
        ModuleInstall('PySocks', 'pip', mname="socks",
                      purpose="A semi-actively maintained SocksiPy fork. Contains many improvements to the original."),
        ModuleInstall('pympler', 'pip',
                      purpose="Pympler is a development tool to measure, monitor and analyze the memory behavior of Python " +
                      "objects in a running Python application."),
        ModuleInstall('fbpca', 'pip',
                      purpose="Functions for principal component analysis (PCA) and accuracy checks."),
        ModuleInstall('prince', 'pip',
                      purpose="Prince is a factor analysis library for datasets that fit in memory."),
        ModuleInstall('pivottablejs', 'pip',
                      purpose="PivotTable.js integration for Jupyter/IPython Notebook."),
        #
        # January 2017
        #
        ModuleInstall('pystan', 'pip', usage="DATA/ML",
                      purpose="PyStan provides an interface to Stan, a package for Bayesian inference using " +
                      "the No-U-Turn sampler, a variant of Hamiltonian Monte Carlo."),
        ModuleInstall('ephem', 'pip', purpose="for fbprophet"),
        ModuleInstall('convertdate', 'pip', purpose="for fbprophet"),
        ModuleInstall('holidays', 'pip', purpose="for fbprophet"),
        ModuleInstall('lunardate', 'pip', purpose="for fbprophet"),
        ModuleInstall('fbprophet', 'pip', usage="DATA/ML",
                      purpose="Prophet is a procedure for forecasting time series data. It is based on an additive " +
                      "model where non-linear trends are fit with yearly and weekly seasonality, plus holidays. " +
                      "It works best with daily periodicity data with at least one year of historical data. " +
                      "Prophet is robust to missing data, shifts in the trend, and large outliers."),
        ModuleInstall('wikipedia', 'pip',
                      purpose="Wikipedia API for Python"),
        ModuleInstall('validate_email', 'pip',
                      purpose="Validate_email verify if an email address is valid and really exists."),
        ModuleInstall('simhash', 'pip',
                      purpose="A Python implementation of Simhash Algorithm"),
        ModuleInstall('wptools', 'pip',
                      purpose="Wikipedia tools (for Humans)"),
        #
        # June 2017
        #
        ModuleInstall('pytest_runner', 'pip',
                      purpose="Setup scripts can use pytest-runner to add setup.py test support for pytest runner."),
        ModuleInstall('fastparquet', 'wheel',
                      purpose="fastparquet is a python implementation of the parquet format, aiming integrate " +
                      "into python-based big data work-flows."),
        ModuleInstall('citeproc-py', 'pip', mname="citeproc_py",
                      purpose="citeproc-py is a CSL processor for Python. It aims to implement the CSL 1.0.1 specification. " +
                      "citeproc-py can output styled citations and bibliographies in a number of different output formats. " +
                      "Currently supported are plain text, reStructuredText and HTML. Other formats can be added easily."),
        ModuleInstall('duecredit', 'pip',
                      purpose="Publications (and donations) tracer"),
        #
        # September 2017
        #
        ModuleInstall('Brotli', 'wheel', mname="brotli",
                      purpose="Brotli is a generic-purpose lossless compression algorithm that compresses data using a " +
                      "combination of a modern variant of the LZ77 algorithm, Huffman coding and 2nd order context modeling, " +
                      "with a compression ratio comparable to the best currently available general-purpose compression methods. " +
                      "It is similar in speed with deflate but offers more dense compression."),
        ModuleInstall('fast-histogram', 'wheel2', mname="fast_histogram",
                      purpose="Mizani is a scales package for graphics. It is written in Python and is " +
                      "based on Hadley Wickham's Scales."),
        ModuleInstall('mizani', 'pip',
                      purpose="Mizani is a scales package for graphics. It is written in Python and is " +
                      "based on Hadley Wickham's Scales."),
        ModuleInstall('mpl-scatter-density', 'pip', mname="mpl_scatter_density",
                      purpose="Matplotlib helpers to make density scatter plots"),
        ModuleInstall('pybind11', 'pip', usage="C++",
                      purpose="pybind11 is a lightweight header-only library that exposes C++ types in Python and vice versa, " +
                      "mainly to create Python bindings of existing C++ code."),
        ModuleInstall('mypy_extensions', 'pip', purpose="for mypy"),
        ModuleInstall('mypy', 'pip',
                      purpose="Mypy is an experimental optional static type checker for Python that aims to combine the benefits of " +
                      "dynamic (or 'duck') typing and static typing."),
        ModuleInstall('pypandoc', 'pip',
                      purpose="Pypandoc provides a thin wrapper for pandoc, a universal document converter."),
        ModuleInstall('pocket', 'pip',
                      purpose="Access to pocket API."),
        #
        # November 2017
        #
        ModuleInstall('seasonal', 'pip', purpose="timeseries", usage="ML"),
        #
        # December 2017
        #
        ModuleInstall('gdown', 'pip',
                      purpose="Google Drive direct download of big files."),
        ModuleInstall('pytube3', 'pip',
                      purpose="play with youtube videos"),
        #
        # Mars, Avril 2018
        #
        ModuleInstall('pyjsparser', 'pip',
                      purpose="Fast javascript parser (based on esprima.js)"),
        ModuleInstall('Js2Py', 'pip', mname="js2py",
                      purpose="JavaScript to Python Translator & JavaScript interpreter written in 100% pure Python."),
        ModuleInstall('pythonnet', 'wheel', mname="clr",
                      purpose="Python binding for C#"),
        #
        # July 2018
        #
        ModuleInstall('lml', 'pip',
                      purpose="lml seamlessly finds the lml based plugins from your current python " +
                      "environment but loads your plugins on demand. It is designed to support plugins " +
                      "that have external dependencies, especially bulky and/or memory hungry ones. " +
                      "lml provides the plugin management system only and the plugin interface is on your shoulder."),
        ModuleInstall('macropy3', 'pip',
                      purpose='MacroPy is an implementation of Syntactic Macros in the Python Programming ' +
                      'Language. MacroPy provides a mechanism for user-defined functions (macros) to perform ' +
                      'transformations on the abstract syntax tree (AST) of a Python program at import time. This ' +
                      'is an easy way to enhance the semantics of a Python program in ways which are otherwise ' +
                      'impossible, for example providing an extremely concise way of declaring classes.'),
        ModuleInstall('dukpy', 'pip',
                      purpose='DukPy is a simple javascript interpreter for Python built on top of duktape ' +
                      'engine without any external dependency. It comes with a bunch of common transpilers ' +
                      'built-in for convenience.'),
        ModuleInstall('javascripthon', 'pip',
                      purpose='a Python 3 to ES6 JavaScript translator'),
        ModuleInstall('pyecharts', 'pip',
                      purpose='pyecharts is a library to generate charts using Echarts. It simply provides ' +
                      'the interface of 28+ kinds of charts between Echarts and Python.'),
        ModuleInstall('pyecharts-javascripthon', 'pip', mname='pyecharts_javascripthon',
                      purpose='pyecharts-javascripthon helps translate Python functions into javascript ones. ' +
                      'It uses javascripthon and dukpy to blend Python codes into javascript runtime. It supports ' +
                      'python 2.7, 3.4, 3.5 and 3.6. It works on Linux, MacOS and Windows platforms.'),
        ModuleInstall('pyecharts-snapshot', 'pip', mname='pyecharts_snapshot',
                      purpose='pyecharts-snapshot renders the output of pyecharts as a png, jpeg, gif, svg image ' +
                      'or a pdf file at command line or in your code.'),
        ModuleInstall('jupyter-echarts-pypkg', 'pip', mname='jupyter_echarts_pypkg',
                      purpose='The project packages jupyter-echarts and distributes it via pypi.'),
        ModuleInstall('yahoo-historical', 'pip', mname="yahoo_historical",
                      purpose='Python module to get stock data from Yahoo! Finance'),
        #
        # October 2019
        #
        ModuleInstall('wsproto', 'pip',
                      purpose='Pure Python, pure state-machine WebSocket implementation.'),
        ModuleInstall('h11', 'pip',
                      purpose='This is a little HTTP/1.1 library written from scratch in Python, '
                              'heavily inspired by hyper-h2.'),
        ModuleInstall('httptools', 'pip',
                      purpose='httptools is a Python binding for nodejs HTTP parser. '
                              'It\'s still in a very early development stage, expect '
                              'APIs to break.'),
        ModuleInstall('hypercorn', 'pip',
                      purpose='Hypercorn is an ASGI web server based on the sans-io hyper, h11, '
                              'h2, and wsproto libraries and inspired by Gunicorn.'),
        ModuleInstall('priority', 'pip',
                      purpose='Priority is a pure-Python implementation of the priority logic '
                              'for HTTP/2.'),
        ModuleInstall('starlette', 'pip',
                      purpose='Starlette is a lightweight ASGI framework/toolkit, which is ideal '
                              'for building high performance asyncio services.'),
        ModuleInstall('uvicorn', 'pip',
                      purpose='Uvicorn is a lightning-fast ASGI server implementation, '
                              'using uvloop and httptools.'),
        ModuleInstall('websockets', 'pip',
                      purpose='websockets is a library for building WebSocket servers and '
                              'clients in Python with a focus on correctness and simplicity.'),
        #
        # December 2019
        #
        ModuleInstall('py-spy', 'pip',
                      purpose='profiler implemented in rust, works like a C++ profiler'),
        ModuleInstall('pydicom', 'pip',
                      purpose='pydicom is a pure python package for working with DICOM files. '
                              'It was made for inspecting and modifying DICOM data in an easy '
                              '"pythonic" way. The modifications can be written again to a new '
                              'file.'),
    ]

    return [_ for _ in mod if _ is not None]



"""
@brief      test tree node (time=3s)
"""

import sys
import os
import unittest
import warnings
from pyquickhelper.loghelper import fLOG
from pyquickhelper.pycode import get_temp_folder
from pymyinstall.installhelper.install_venv_helper import create_virtual_env, run_venv_script


class TestInstallWithDeps(unittest.TestCase):

    def test_venv_install_deps(self):
        fLOG(
            __file__,
            self._testMethodName,
            OutputPrint=__name__ == "__main__")

        fold = os.path.dirname(os.path.abspath(__file__))
        src_ = os.path.normpath(os.path.abspath(
            os.path.join(fold, "..", "..", "src")))
        assert os.path.exists(src_)
        temp = get_temp_folder(__file__, "temp_venv_install_deps")

        if __name__ != "__main__":
            warnings.warn("does not work well from a virtual environment")
            return

        if sys.version_info[0] == 2:
            # using nose so previous test is true
            warnings.warn(
                "does not work well from a virtual environment (Python 2.7)")
            return

        out = create_virtual_env(temp, fLOG=fLOG)
        src_ = src_.replace("\\", "/")
        temp = temp.replace("\\", "/")

        script = ["import sys",
                  "sys.path.append('{0}')".format(src_),
                  "from pymyinstall.packaged import install_module_deps",
                  "install_module_deps('imbox', temp_folder='{0}', source='2')".format(
                      temp),
                  ]

        file_script = os.path.join(temp, "test_install_deps.py")
        with open(file_script, "w") as f:
            f.write("\n".join(script))

        out = run_venv_script(temp, file_script, fLOG=fLOG, file=True)
        if "installing module  six" not in out:
            raise Exception(out)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
@file
@brief Defines a set of modules to try
"""

from ..installhelper.module_install import ModuleInstall


def scraping_set():
    """
    modules which help scraping the web, it requires the modules in set *small*
    """
    mod = [
        ModuleInstall(
            "langdetect", "pip", usage="WEB",
            purpose="Language detection library ported from Google's language-detection."),
        ModuleInstall(
            "justext", "pip", usage="WEB",
            purpose="Program jusText is a tool for removing boilerplate content, such as navigation links, headers, " +
            "and footers from HTML pages. It is designed to preserve mainly text containing full " +
            "sentences and it is therefore well suited for creating linguistic resources such as Web corpora."),
        ModuleInstall(
            "tldextract", "pip", usage="WEB",
            purpose="Accurately separate the TLD from the registered domain and subdomains of a URL, using the Public Suffix List."),
        ModuleInstall(
            "cchardet", "wheel", usage="WEB",
            purpose="Universal encoding detector. This library is faster than chardet."),
        ModuleInstall(
            "multidict", "pip",
            purpose="Multidicts are useful for working with HTTP headers, URL query args etc."),
        ModuleInstall(
            "async_timeout", "pip", usage="WEB",
            purpose="Timeout context manager for asyncio programs"),
        ModuleInstall(
            "yarl", "pip", usage="WEB",
            purpose="Yet another URL library"),
        ModuleInstall(
            "idna_ssl", "pip", usage="WEB",
            purpose="Patch ssl.match_hostname for Unicode(idna) domains support"),
        ModuleInstall(
            "aiohttp", "wheel", usage="WEB",
            purpose="http client/server for asyncio"),
        ModuleInstall(
            "sky", "pip", usage="WEB",
            purpose="sky is a web scraping framework, implemented with the latest python versions in mind (3.4+). " +
            "It uses the asynchronous asyncio framework, " +
            "as well as many popular modules and extensions."),
    ]

    return [_ for _ in mod if _ is not None]



"""
@file
@brief Functions to get module version, license, dependencies
"""
import sys
import re
import warnings
import functools
import time
import platform
import xmlrpc.client as xmlrpc_client
from .install_cmd_helper import run_cmd, get_pip_program
from .module_install_exceptions import (
    MissingPackageOnPyPiException, AnnoyingPackageException)
from .module_install_exceptions import (
    ConfigurationError, MissingVersionOnPyPiException,
    WrongVersionError, MissingVersionWheelException)
from .install_cmd_regex import regex_wheel_versions
from .pip_helper import get_installed_distributions


annoying_modules = {"pygame", "liblinear", "mlpy", "VideoCapture",
                    "libsvm", "opencv_python", "scikits.cuda",
                    "NLopt"}


_get_module_version_manual_memoize = {}


def get_module_version(module, use_cmd=False):
    """
    Returns a dictionary ``{ module: version }``.

    @param      module      unused, None
    @param      use_cmd     use command line
    @return                 dictionary
    """
    if module is not None:
        modl = module.lower()
        res = get_module_version(None, use_cmd=use_cmd)
        return res.get(modl, None)

    global _get_module_version_manual_memoize  # pylint: disable=W0602
    if len(_get_module_version_manual_memoize) > 0:
        return _get_module_version_manual_memoize

    res = {}

    if use_cmd:
        prog = get_pip_program()
        cmd = prog + " list"
        out, err = run_cmd(cmd, wait=True, fLOG=None)

        if err is not None and len(err) > 0:
            if len(err.split("\n")) > 3 or \
               "You should consider upgrading via the 'pip install --upgrade pip' command." not in err:
                raise Exception("unable to run, #lines {0}\nCMD:\n{3}\nERR-J:\n{1}\nOUT:\n{2}".format(
                    len(err.split("\n")), err, out, cmd))
        lines = out.split("\n")

        for line in lines:
            if "(" in line:
                spl = line.split()
                if len(spl) == 2:
                    a = spl[0]
                    b = spl[1].strip(" \n\r")
                    res[a] = b.strip("()")
                    al = a.lower()
                    if al != a:
                        res[al] = res[a]
    else:
        # local_only must be False to get all modules
        # not only the ones installed in the virtual environment
        dist = get_installed_distributions(local_only=False)
        if len(dist) == 0:
            raise ConfigurationError("no installed module, unexpected (pip should be there): " +
                                     "sys.executable={0}, sys.prefix={1}, sys.base_prefix={2}".format(
                                         sys.executable, sys.prefix, sys.base_prefix))
        for mod in dist:
            al = mod.key.lower()
            a = mod.key
            try:
                v = mod.version
            except ValueError:
                v = "UNKNOWN"
            res[a] = v
            if a != al:
                res[al] = v

    _get_module_version_manual_memoize.update(res)
    return res


def is_installed(name):
    """
    Tells if a module is installed or not.

    @param      name        module name
    @return                 boolean
    """
    return get_module_version(name) is not None


_get_module_metadata_manual_memoize = {}


def get_module_metadata(module, use_cmd=False, refresh_cache=False):
    """
    Returns a dictionary ``{ module: metadata }``.

    @param      module          unused, None
    @param      refresh_cache   refresh the cache before getting metadata
    @return                     dictionary
    """
    if module is not None:
        modl = module.lower()
        res = get_module_metadata(
            None, use_cmd=use_cmd, refresh_cache=refresh_cache)
        return res.get(modl, None)

    global _get_module_metadata_manual_memoize  # pylint: disable=W0602
    if not refresh_cache and len(_get_module_metadata_manual_memoize) > 0:
        return _get_module_metadata_manual_memoize

    # local_only must be False to get all modules
    # not only the ones installed in the virtual environment
    dist = get_installed_distributions(local_only=False, use_cmd=use_cmd)
    if len(dist) == 0:
        raise ConfigurationError("no installed module, unexpected (pip should be there): " +
                                 "sys.executable={0}, sys.prefix={1}, sys.base_prefix={2}".format(
                                     sys.executable, sys.prefix, sys.base_prefix))
    res = {}
    for mod in dist:
        d = {}
        lines = mod._get_metadata(mod.PKG_INFO)
        for line in lines:
            if sys.version_info[0] == 2:
                typstr = str  # unicode#
                if not isinstance(line, typstr):
                    line = typstr(line, encoding="utf8", errors="ignore")
            try:
                spl = line.split(":")
            except UnicodeDecodeError:
                warnings.warn("UnicodeDecodeError with: " + line)
                continue
            key = spl[0].strip()
            value = ":".join(spl[1:]).strip()
            if key not in d:
                d[key] = value
            else:
                if not isinstance(d[key], list):
                    d[key] = [d[key]]
                d[key].append(value)

        a = mod.key
        res[a] = d
        al = mod.key.lower()
        if a != al:
            res[al] = d

    _get_module_metadata_manual_memoize.update(res)
    return res


def _get_pypi_version_memoize_op(f):
    memo = {}

    def helper(module_name, full_list=False, url="https://pypi.python.org/pypi"):
        key = module_name, full_list, url
        if key not in memo:
            memo[key] = f(module_name, full_list, url)
        return memo[key]
    return helper


_get_pypi_version_memoize = {}


def get_pypi_version(module_name, full_list=False, url="https://pypi.python.org/pypi", skip_betas=True):
    """
    Returns the version of a package on :epkg:`pypi`,
    we skip alpha, beta or dev version.

    @param      module_name     module name
    @param      url             pypi server
    @param      full_list       results as a list or return the last stable version
    @param      skip_betas      skip the intermediate functions
    @return                     version (str or list)

    See also `installing_python_packages_programatically.py <https://gist.github.com/rwilcox/755524>`_,
    `pkgtools.pypi: PyPI interface <http://pkgtools.readthedocs.org/en/latest/pypi.html>`_.

    It the function fails, check the status of
    `Python Infrastructure <https://status.python.org/>`_.
    It can return errors::

        ProtocolError: ProtocolError for pypi.python.org/pypi: 503 No healthy backends
    """

    global _get_pypi_version_memoize  # pylint: disable=W0602
    key = module_name, full_list, url
    if key in _get_pypi_version_memoize:
        available = _get_pypi_version_memoize[key]
        if full_list:
            return available
        elif available is not None and len(available) > 0:
            return available[0]
        return None

    def pypi_package_releases(module_name, b):
        nbtry = 0
        while nbtry < 40:
            try:
                available = pypi.package_releases(module_name, True)
                return available
            except Exception as e:
                if ("HTTPTooManyRequests" in str(type(e)) or
                        "HTTPTooManyRequests" in str(e)):
                    nbtry += 1
                    warnings.warn(str(e))
                    time.sleep(90)
                    continue
                if ("TimeoutError" in str(type(e)) or
                        "TimeoutError" in str(e)):
                    nbtry += 1
                    warnings.warn(str(e))
                    time.sleep(20)
                    continue
                raise e
        return None

    def _inside_loop_(pypi, module_name, tried):

        available = pypi_package_releases(module_name, True)

        if available is None or len(available) == 0:
            ntry = module_name.capitalize()
            if ntry not in tried:
                tried.append(ntry)
                available = pypi_package_releases(tried[-1], True)

        if available is None or len(available) == 0:
            ntry = module_name.replace("-", "_")
            if ntry not in tried:
                tried.append(ntry)
                available = pypi_package_releases(tried[-1], True)

        if available is None or len(available) == 0:
            ntry = module_name.replace("_", "-")
            if ntry not in tried:
                tried.append(ntry)
                available = pypi_package_releases(tried[-1], True)

        if available is None or len(available) == 0:
            ntry = module_name.lower()
            if ntry not in tried:
                tried.append(ntry)
                available = pypi_package_releases(tried[-1], True)

        if available is None or len(available) == 0:
            ml = module_name.lower()
            if ml == "markupsafe":
                tried.append("MarkupSafe")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "flask-sqlalchemy":
                tried.append("Flask-SQLAlchemy")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "apscheduler":
                tried.append("APScheduler")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "datashape":
                tried.append("DataShape")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "pycontracts":
                tried.append("PyContracts")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "pybrain":
                tried.append("PyBrain")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "pyexecjs":
                tried.append("PyExecJS")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "dataspyre":
                tried.append("DataSpyre")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "heapdict":
                tried.append("HeapDict")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "pyreact":
                tried.append("PyReact")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "qtpy":
                tried.append("QtPy")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "pythonqwt":
                tried.append("PythonQwt")
                available = pypi_package_releases(tried[-1], True)
            elif ml == "onedrive-sdk-python":
                tried.append("onedrivesdk")
                available = pypi_package_releases(tried[-1], True)
            elif ml.startswith("orange3-"):
                s = ml.split("-")[1]
                ntry = "Orange3-" + s[0].upper() + s[1:]
                tried.append(ntry)
                available = pypi_package_releases(tried[-1], True)
            elif module_name in annoying_modules:
                raise AnnoyingPackageException(module_name)

        # this raises a warning about an opened connection
        # see documentation of the function
        # del pypi

        return available

    tried = [module_name]

    if sys.version_info[:2] <= (3, 4):
        # the client does not an implemented of __exit__ for version <= 3.4
        pypi = xmlrpc_client.ServerProxy(url)
        available = _inside_loop_(pypi, module_name, tried)
    else:
        with xmlrpc_client.ServerProxy(url) as pypi:
            available = _inside_loop_(pypi, module_name, tried)

    if available is None or len(available) == 0:
        raise MissingPackageOnPyPiException("tried:\n" + "\n".join(tried))

    def filter_betas(a):
        spl = a.split(".")
        if len(spl) in (2, 3):
            last = spl[-1]
            if not skip_betas or ("a" not in last and "b" not in last and "dev" not in last):
                return True
        else:
            # we don't really know here, so we assume it is not
            return True
        return False

    if available:
        available2 = list(filter(filter_betas, available))
        if available2:
            _get_pypi_version_memoize[key] = available2
            if full_list:
                return available2
            return available2[0]

    raise MissingVersionOnPyPiException(
        "{0}\nversion:\n{1}".format(module_name, "\n".join(available)))


def numeric_version(vers):
    """
    convert a string into a tuple with numbers wherever possible

    @param      vers    string
    @return             tuple
    """
    if isinstance(vers, tuple):
        return vers
    if isinstance(vers, list):
        raise Exception("unexpected value:" + str(vers))
    spl = str(vers).split(".")
    r = []
    for _ in spl:
        try:
            i = int(_)
            r.append(i)
        except ValueError:
            r.append(_)
    return tuple(r)


def compare_version(num, vers):
    """
    Compares two versions.

    @param      num     first version
    @param      vers    second version
    @return             -1, 0, 1
    """
    if num is None:
        if vers is None:
            return 0
        return 1
    if vers is None:
        return -1

    if not isinstance(vers, tuple):
        vers = numeric_version(vers)
    if not isinstance(num, tuple):
        num = numeric_version(num)

    if len(num) == len(vers):
        for a, b in zip(num, vers):
            if isinstance(a, int) and isinstance(b, int):
                if a < b:
                    return -1
                if a > b:
                    return 1
            else:
                a = str(a)
                b = str(b)
                if a < b:
                    return -1
                if a > b:
                    return 1
        return 0

    if len(num) < len(vers):
        num = num + (0,) * (len(vers) - len(num))
        return compare_version(num, vers)
    vers = vers + (0,) * (len(num) - len(vers))
    return compare_version(num, vers)


def version_consensus(v1, v2):
    """
    *v1* and *v2* are two versions of the same module, which one to keep?

    @param      v1      version 1
    @param      v2      version 2
    @return             consensus

    * ``v1=None``, ``v2='(>=1.5)'`` --> ``v='>=1.5'``

    To improve:

    * ``v1='<=1.6'``, ``v2='(>=1.5)'`` --> ``v='==1.6'``
    """
    reg = re.compile("([><=!]*)([^><=!]+)")

    def process_version(v):
        if isinstance(v, str  # unicode#
                      ):
            v = v.strip('()')
            find = reg.search(v)
            if not find:
                raise WrongVersionError(v)
            sign = find.groups()[0]
            number = numeric_version(find.groups()[1])
        else:
            try:
                sign, number = v
            except ValueError as e:
                raise ValueError("weird format: " + str(v) +
                                 " - " + str(type(v))) from e
        return sign, number

    if v1 is None:
        return v2
    if v2 is None:
        return v1

    s1, n1 = process_version(v1)
    s2, n2 = process_version(v2)

    if s1 not in ('<=', '==', '<', '>', '>=', '!='):
        raise ValueError("wrong sign '{0}' for v1='{1}'".format(s1, v1))
    if s2 not in ('<=', '==', '<', '>', '>=', '!='):
        raise ValueError("wrong sign '{0}' for v1='{1}'".format(s2, v2))

    if s1 == "==":
        if s2 == "==":
            if compare_version(n1, n2) != 0:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s1, n1
        else:
            res = s1, n1

    elif s1 == "<=":
        if s2 == "<=":
            res = s1, min(n1, n2)
        elif s2 == "==":
            if compare_version(n1, n2) < 0:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s2, n2
        elif s2 == '<':
            if compare_version(n1, n2) == -1:
                res = s1, n1
            else:
                res = s2, n2
        elif s2 in ('>', '>='):
            if compare_version(n1, n2) <= 0:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s1, n1

    elif s1 == "<":
        if s2 == "<":
            res = s1, min(n1, n2)
        elif s2 == "==":
            if compare_version(n1, n2) <= 0:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s2, n2
        elif s2 == '<=':
            if compare_version(n1, n2) <= 0:
                res = s1, n1
            else:
                res = s2, n2
        elif s2 in ('>', '>='):
            if compare_version(n1, n2) <= 0:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s1, n1

    elif s1 == ">=":
        if s2 == ">=":
            res = s1, max(n1, n2)
        elif s2 == "==":
            if compare_version(n1, n2) == -1:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s2, n2
        elif s2 == '>':
            if compare_version(n1, n2) <= 0:
                res = s2, n2
            else:
                res = s1, n1
        elif s2 in ('<', '<='):
            if compare_version(n1, n2) >= 0:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s2, n2
        elif s2 == '!=':
            if compare_version(n1, n2) == 0:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s1, max(n1, n2)

    elif s1 == ">":
        if s2 == ">":
            res = s1, max(n1, n2)
        elif s2 == "==":
            if compare_version(n1, n2) >= 0:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s2, n2
        elif s2 == '>=':
            if compare_version(n1, n2) == -1:
                res = s2, n2
            else:
                res = s1, n1
        elif s2 in ('<', '<='):
            if compare_version(n1, n2) == 1:
                raise WrongVersionError(
                    "incompatible version: {0}{1} and {2}{3}".format(s1, n1, s2, n2))
            res = s2, n2
    else:
        res = None, None

    if res[0] is None:
        raise WrongVersionError(
            "incompatible version and wrong format: {0}{1} and {2}{3}".format(s1, n1, s2, n2))

    return '{0}{1}'.format(res[0], '.'.join(str(_) for _ in res[1]))


_get_module_dependencies_deps = None


def get_module_dependencies(module, use_cmd=False, deep=False, collapse=True, use_pip=None, refresh_cache=False):
    """
    Returns the dependencies for a module.

    @param      module          unused, None
    @param      use_cmd         use command line
    @param      deep            dig into dependencies of dependencies
    @param      collapse        only one row per module
    @param      use_pip         use pip to discover dependencies or not (parse metadata)
    @param      refresh_cache   refresh the cache (see below)
    @return                     list of tuple (module, version, required by as a str)
                                or dictionary  { module: (version, required by as a list) } if *collapse* is True

    The function which uses *use_pip=True* is not fully tested, it does not
    return contraints (== 2.4). The function caches the results to avoid doing it again
    during a second execution unless *refresh_cache* is True.
    This function is not tested on Python 2.7.
    """
    if use_pip is None:
        use_pip = not sys.platform.startswith("win")

    def evaluate_condition(cond, full):
        python_version = ".".join(str(_) for _ in sys.version_info[:3])
        extra = ""
        sys_platform = sys.platform
        platform_machine = platform.machine
        platform_python_implementation = platform.python_implementation
        try:
            return eval(cond)
        except Exception:
            if "python_version" not in cond and "extra" not in cond:
                # probably something like cycler (>=0.10)
                # we don't check that
                return True
            mes = ("Unable to evaluate condition '{0}' from '{1}', "
                   "extra='{2}', python_version='{3}', "
                   "sys_platform='{4}' platform_machine='{5}', "
                   "platform_python_implementation='{6}'.").format(
                cond, full, extra, python_version, sys_platform,
                platform_machine, platform_python_implementation)
            raise Exception(mes)

    if use_pip:
        global _get_module_dependencies_deps
        if _get_module_dependencies_deps is None or refresh_cache:
            temp = get_installed_distributions(
                local_only=False, skip=[], use_cmd=use_cmd)
            _get_module_dependencies_deps = dict(
                (p.key, (p, p.requires())) for p in temp)
        if module not in _get_module_dependencies_deps:
            raise ValueError("module {0} was not installed".format(module))
        res = []
        req = _get_module_dependencies_deps[module][1]
        if isinstance(req, list):
            for r in req:
                res.append((r.key, None, module))
        else:
            res.append((req.key, None, module))
    else:
        meta = get_module_metadata(
            module, use_cmd, refresh_cache=refresh_cache)
        if meta is None:
            raise ImportError(
                "unable to get metadata for module '{0}' - refresh_cache={1}".format(module, refresh_cache))
        deps = [v for k, v in meta.items()
                if "Requires" in k and "Requires-Python" not in k]
        res = []
        for d in deps:
            if not isinstance(d, list):
                dl = [d]
            else:
                dl = d
            for v in dl:
                spl = v.split()
                if len(spl) > 1:
                    spl = [spl[0], " ".join(spl[1:])]
                if len(spl) <= 1:
                    key = (module, None, v)
                else:
                    conds = spl[1].split(";")
                    ok = [evaluate_condition(cond, v) for cond in conds]
                    if not all(ok):
                        continue
                    key = (spl[0].strip(";"), spl[1], module)
                if key not in res:
                    res.append(key)

    # specific filters
    def validate_module(name):
        if name == "enum34" and sys.version_info[:2] > (3, ):
            raise NameError(
                "Unexpected dependency '{0}' for module '{1}'.".format(name, module))
        if name == "configparser":
            raise NameError(
                "Unexpected dependency '{0}' for module '{1}'.".format(name, module))
        return True

    res = [key for key in res if validate_module(key[0])]

    if deep:
        done = {module: None}
        mod = 1
        while mod > 0:
            mod = 0
            for r in res:
                if r[0] not in done:
                    if r[0].lower() < 'a' or r[0].lower() > 'z' or r[0].endswith(";"):
                        raise NameError(
                            "A module has an unexpected name '{0}', r={1} "
                            "when looking for dependencies of '{2}'.".format(
                                r[0], r, module))
                    temp = get_module_dependencies(
                        r[0], use_cmd=use_cmd, deep=deep, collapse=False,
                        use_pip=use_pip, refresh_cache=refresh_cache)
                    for key in temp:
                        if key not in res:
                            res.append(key)
                    mod += 1
                    done[r[0]] = None

    if collapse:
        final = {}
        for name, version, required in res:
            if name not in final:
                final[name] = (version, [required])
            else:
                ex = final[name][1]
                if required not in ex:
                    ex.append(required)
                    try:
                        v = version_consensus(final[name][0], version)
                    except WrongVersionError as e:
                        raise WrongVersionError("unable to reconcile versions:\n{0}\n{1}".format(
                            ex, str((name, version, required)))) from e
                    final[name] = (v, ex)
        final = {k: (v[0], list(sorted(v[1]))) for k, v in final.items()}
        return final

    return [(name, version.strip('()') if version is not None else version, required)
            for name, version, required in res]


def choose_most_recent(list_name):
    """
    Chooses the most recent version for a list of module names.

    @param      list_name       list of names
    @return                     most recent version or None if the input list is empty

    In the following case, we would choose the first option::

        numpy-1.10.0+mkl-cp34-none-win_amd64.whl
        numpy-1.9.1.0+mkl-cp34-none-win_amd64.whl
    """
    def find_wheel(tu):
        for t in tu:
            if ".whl" in t:
                return t
        raise ValueError("unable to find a wheel in {0}".format(tu))
    if len(list_name) == 0:
        return None
    if isinstance(list_name[0], tuple):
        list_name = [(find_wheel(_), _) for _ in list_name]
    else:
        list_name = [(_, _) for _ in list_name]

    versions = [re.compile(_) for _ in regex_wheel_versions]

    def search_regex(_):
        resv = None
        for version in versions:
            try:
                resv = version.search(_[0])
            except TypeError as e:
                raise TypeError("Unable to parse '{0}'".format(_)) from e
            if resv is not None:
                return resv
        raise MissingVersionWheelException(
            "Unable to get version number for module '{}':\nREGEX\n{}".format(_, "\n".join(regex_wheel_versions)))

    list_name = [(search_regex(_).groups()[0], _[0], _[1])
                 for _ in list_name]

    def cmp(el1, el2):
        return compare_version(el1[0], el2[0])

    list_name = list(sorted(list_name, key=functools.cmp_to_key(cmp)))
    return list_name[-1][-1]


def get_wheel_version(whlname):
    """
    extract the version from a wheel file,
    return ``2.6.0`` for ``rpy2-2.6.0-cp34-none-win_amd64.whl``

    @param      whlname     file name
    @return                 string
    """
    find = []
    for reg in regex_wheel_versions:
        if len(find) == 0:
            exp = re.compile(reg)
            find = exp.findall(whlname)
        else:
            break
    if len(find) == 0:
        mes = "Unable to extract version for '{0}'\nREGEX\n{1}"
        raise ValueError(mes.format(whlname, "\n".join(regex_wheel_versions)))
    if len(find) > 1:
        mes = "Too many options for '{0}'\nOPTIONS\n{1}\nREGEX\n{2}"
        raise ValueError(mes.format(
            whlname, find, "\n".join(regex_wheel_versions)))
    if isinstance(find[0], tuple):
        return find[0][0]
    else:
        return find[0]



"""
@file
@breif shortcuts to win_installer
"""

from .pywin32_helper import import_pywin32, fix_pywin32_installation
from .win_exception import WinInstallException
from .win_extract import extract_msi, extract_exe
from .win_setup_r import r_run_script
from .win_setup_main import win_python_setup
from .win_innosetup_helper import inno_install_kernels
from .win_packages import win_install_package_other_python, is_package_installed
from .win_patch import win_patch_paths
from .win_setup_main_checkings import distribution_checkings, import_every_module



"""
@file
@brief Various functions to install some application such as `scala-sbt <http://www.scala-sbt.org/download.html>`_.
"""
from __future__ import print_function
import sys
import re
import os

from ..installhelper.install_cmd_helper import run_cmd
from .install_custom import download_page, download_file


def install_scala_sbt(
        temp_folder=".", fLOG=print, install=True, force_download=False, version=None):
    """
    Install `scala-sbt <http://www.scala-sbt.org/download.html>`_.
    It does not do it a second time if it is already installed.

    @param      temp_folder     where to download the setup
    @param      fLOG            logging function
    @param      install         install (otherwise only download)
    @param      force_download  force the downloading of pandoc
    @param      version         version to install (unused)
    @return                     temporary file
    """
    if version is not None:
        raise ValueError("cannot specify a version")
    #bb = IsPandocInstalled()
    bb = False
    if bb and not force_download:
        return True

    link = "http://www.scala-sbt.org/download.html"
    page = download_page(link)
    if sys.platform.startswith("win"):
        reg = re.compile("href=\\\"(https://.*?msi)\\\"")
        alls = reg.findall(page)
        if len(alls) == 0:
            raise Exception(
                "unable to find a link on a .msi file on page: " +
                page)

        file = alls[0]
        filel = file
        full = filel
        version = file.split("/")[-2]
        fLOG("[pymy] scala-sbt, version ", version)
        outfile = os.path.join(temp_folder, full.split("/")[-1])
        if not outfile.endswith(".msi"):
            outfile += ".msi"
        fLOG("[pymy] download ", full)
        local = download_file(full, outfile)
        if install and not bb:
            run_cmd("msiexec /i " + local, fLOG=fLOG, wait=True)
        return local
    else:
        raise NotImplementedError("not available on platform " + sys.platform)



import sys
import os
sys.path.append("src")
from pymyinstall.win_installer import import_pywin32, fix_pywin32_installation

try:
    import win32com
except ImportError as e:
    fix_pywin32_installation()

#exe = os.path.abspath(os.path.dirname(sys.executable))
#os.environ["PATH"] = os.environ["PATH"] + ";" + exe

import win32com
